#! /usr/bin/env python3

"""Test datastore module."""

import unittest

import os
import inspect
import json
import time

from h5py._hl.files import File

from qmi.data.datastore import DataFolder, DataStore
from qmi.core.config_defs import CfgLogging
from qmi.data.dataset import DataSet
import qmi.core.exceptions

def _create_config(config_struct_class):
    config = {item[0]: item[1] for item in inspect.getmembers(config_struct_class()) if "__" not in item[0]}
    return config

def _create_dataset():
    ds = DataSet("my_dataset", shape=(2, 8, 3))

    ds.set_axis_label(0, "X")
    ds.set_axis_unit(0, "um")
    ds.set_axis_label(1, "Z")
    ds.set_axis_unit(1, "mm")
    ds.set_column_label(0, "power")
    ds.set_column_unit(0, "mW")
    ds.set_column_label(1, "countrate")
    ds.set_column_unit(1, "kHz")
    ds.set_column_label(2, "temperature")
    ds.set_column_unit(2, "K")
    
    return ds


class TestDataFolder(unittest.TestCase):

    def setUp(self) -> None:
        self.datafolder = DataFolder(os.getcwd(), "test", None, None)

    def tearDown(self) -> None:
        del self.datafolder

    def test_01_create_config_json_file(self):
        """Create a JSON file from given input `configstruct` class"""
        # Arrange
        expected_config = _create_config(CfgLogging)
        # Act
        self.datafolder.write_config(CfgLogging())
        # Assert
        try:
            with open(os.path.join(os.getcwd(), "test.json")) as f:
                written_config = json.load(f)

            self.assertDictEqual(expected_config, written_config)

        finally:
            os.remove(os.path.join(os.getcwd(), "test.json"))

    def test_02_make_file_copy(self):
        """Make a copy of a file"""
        # Arrange
        name = "test.file"
        expected_copied_file = os.path.join(os.getcwd(), name)
        tmp_folder = os.path.join(os.getcwd(), "tmp")
        os.mkdir(tmp_folder)
        file_to_copy = os.path.join(tmp_folder, name)
        try:
            with open(file_to_copy, "w") as f:
                f.write("bla")

            # act
            self.datafolder.copy_file(file_to_copy)

            # assert
            self.assertTrue(os.path.isfile(expected_copied_file))

        finally:
            os.remove(file_to_copy)
            os.removedirs(tmp_folder)
            os.remove(expected_copied_file)

    def test_03_write_dataset_to_hdf5(self):
        """Write a data set as HDF5 file.
        """
        # Arrange
        dataset = _create_dataset()
        expected_file = os.path.join(os.getcwd(), dataset.name + ".h5")
        # Act
        try:
            self.datafolder.write_dataset(dataset)
            # Assert
            self.assertTrue(os.path.isfile(expected_file))

        finally:
            os.remove(expected_file)

    def test_04_write_dataset_to_dat(self):
        """Write a data set as DAT file.
        """
        # Arrange
        dataset = _create_dataset()
        expected_file = os.path.join(os.getcwd(), dataset.name + ".dat")
        # Act
        try:
            self.datafolder.write_dataset(dataset, file_format="text")
            # Assert
            self.assertTrue(os.path.isfile(expected_file))

        finally:
            os.remove(expected_file)

    def test_05_write_dataset_wrong_file_format_raises_exception(self):
        """Wrong file format string raises an expection
        """
        # Arrange
        dataset = _create_dataset()
        # Act and assert
        with self.assertRaises(qmi.core.exceptions.QMI_UsageException):
            self.datafolder.write_dataset(dataset, file_format="boh")

    def test_06_read_dataset_in_hdf5(self):
        """See that we can read in a data set in HDF5 format"""
        # Arrange
        expected_dataset = _create_dataset()
        expected_file = os.path.join(os.getcwd(), expected_dataset.name + ".h5")
        try:
            self.datafolder.write_dataset(expected_dataset)
            # Act
            dataset = self.datafolder.read_dataset(expected_dataset.name)
            # Assert
            self.assertEqual(expected_dataset.name, dataset.name)
            self.assertDictEqual(expected_dataset.attrs, dataset.attrs)
            self.assertListEqual(expected_dataset.axis_label, dataset.axis_label)
            self.assertListEqual(expected_dataset.axis_scale, dataset.axis_scale)
            self.assertListEqual(expected_dataset.axis_unit, dataset.axis_unit)
            self.assertListEqual(expected_dataset.column_label, dataset.column_label)
            self.assertListEqual(expected_dataset.column_unit, dataset.column_unit)
            self.assertEqual(expected_dataset.data.min(), dataset.data.min())
            self.assertEqual(expected_dataset.data.max(), dataset.data.max())
            self.assertEqual(expected_dataset.data.size, dataset.data.size)
            self.assertEqual(expected_dataset.timestamp, dataset.timestamp)

        finally:
            os.remove(expected_file)

    def test_07_read_dataset_in_dat(self):
        """See that we can read in a data set in DAT format"""
        # Arrange
        expected_dataset = _create_dataset()
        expected_file = os.path.join(os.getcwd(), expected_dataset.name + ".dat")
        try:
            self.datafolder.write_dataset(expected_dataset, file_format="text")
            # Act
            dataset = self.datafolder.read_dataset(expected_dataset.name)
            # Assert
            self.assertEqual(expected_dataset.name, dataset.name)
            self.assertDictEqual(expected_dataset.attrs, dataset.attrs)
            self.assertListEqual(expected_dataset.axis_label, dataset.axis_label)
            self.assertListEqual(expected_dataset.axis_scale, dataset.axis_scale)
            self.assertListEqual(expected_dataset.axis_unit, dataset.axis_unit)
            self.assertListEqual(expected_dataset.column_label, dataset.column_label)
            self.assertListEqual(expected_dataset.column_unit, dataset.column_unit)
            self.assertEqual(expected_dataset.data.min(), dataset.data.min())
            self.assertEqual(expected_dataset.data.max(), dataset.data.max())
            self.assertEqual(expected_dataset.data.size, dataset.data.size)
            self.assertEqual(expected_dataset.timestamp, dataset.timestamp)

        finally:
            os.remove(expected_file)

    def test_08_read_dataset_wrong_file_name_raises_exception(self):
        """Wrong file name string raises an expection
        """
        # Act and assert
        with self.assertRaises(FileNotFoundError):
            self.datafolder.read_dataset("boh")

    def test_09_make_hdf5_file(self):
        """Make a hdf5 file"""
        # Arrange
        name = "expected"
        expected_file = os.path.join(os.getcwd(), name + ".h5")
        # Act
        try:
            with self.datafolder.make_hdf5file(name) as hdf5_file:
                self.assertTrue(os.path.isfile(expected_file))
                self.assertEqual(type(hdf5_file), File)

        finally:
            os.remove(expected_file)

    def test_10_make_hdf5_file_raises_exception_with_non_latin_characters(self):
        """Trying to make a file with non-latin characters raises and exception"""
        # Arrange
        name = "väärin"
        # Act and Assert
        with self.assertRaises(ValueError):
            self.datafolder.make_hdf5file(name)

    def test_11_open_hdf5_file(self):
        """Open a hdf5 file"""
        # Arrange
        name = "expected"
        expected_file = os.path.join(os.getcwd(), name + ".h5")
        # Act and Assert
        try:
            with self.datafolder.make_hdf5file(name):
                self.assertTrue(os.path.isfile(expected_file))

            with self.datafolder.open_hdf5file(name) as hdf5_file:
                self.assertEqual(type(hdf5_file), File)

        finally:
            os.remove(expected_file)

    def test_12_open_hdf5_file_raises_exception_with_non_latin_characters(self):
        """Opening a file with non-latin characters raises and exception"""
        # Arrange
        name = "väärin"
        # Act and Assert
        with self.assertRaises(ValueError):
            self.datafolder.open_hdf5file(name)

    def test_13_write_dataset_again(self):
        """Write a data set if already exists.
        """
        # Arrange
        dataset = _create_dataset()
        expected_file = os.path.join(os.getcwd(), dataset.name + ".h5")
        self.datafolder.write_dataset(dataset)
        # Act
        try:
            # Assert
            with self.assertRaises(FileExistsError):
                self.datafolder.write_dataset(dataset)
            self.assertTrue(os.path.isfile(expected_file))
        finally:
            os.remove(expected_file)

    def test_14_write_dataset_again(self):
        """Write a data set if already exists, but with overwrite flag set.
        """
        # Arrange
        dataset = _create_dataset()
        expected_file = os.path.join(os.getcwd(), dataset.name + ".h5")
        self.datafolder.write_dataset(dataset)
        # Act
        try:
            self.datafolder.write_dataset(dataset, overwrite=True)
            # Assert
            self.assertTrue(os.path.isfile(expected_file))
        finally:
            os.remove(expected_file)


class TestDataFolderNoLabel(unittest.TestCase):

    def setUp(self) -> None:
        self.datafolder = DataFolder(os.getcwd(), None, None, None)

    def tearDown(self) -> None:
        del self.datafolder

    def test_01_create_config_json_file(self):
        """Create a JSON file from given input `configstruct` class with default name 'config.json' if no label
        is given.
        """
        # Arrange
        expected_file = os.path.join(os.getcwd(), "config.json")
        # Act
        self.datafolder.write_config(CfgLogging())
        # Assert
        try:
            self.assertTrue(os.path.isfile(expected_file))

        finally:
            os.remove(expected_file)


class TestDataStore(unittest.TestCase):

    def setUp(self) -> None:
        self.datastore = DataStore(os.getcwd())

    def tearDown(self) -> None:
        del self.datastore

    def test_01_make_folder_only_name_as_input(self):
        """Make a new datastore folder in the base folder and that a DataFolder instance is returned."""
        # Arrange
        name = "temp"
        tm = time.localtime(time.time())
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        day_folder = os.path.join(os.getcwd(), date_str)
        expected_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
        try:
            # Act
            datafolder = self.datastore.make_folder(name)
            # We need to check for the rare case that the second just flipped between arranging the test and making
            # the folder. We ignore the super-rare case of this second being exactly at midnight, changing date also
            if not os.path.isdir(expected_folder):
                # Update to 'next' second
                tm = time.localtime(time.time())
                time_str = time.strftime("%H%M%S", tm)
                expected_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))

            # Assert
            self.assertTrue(os.path.isdir(expected_folder))
            self.assertEqual(type(datafolder), DataFolder)

        finally:
            os.removedirs(expected_folder)
            # os.removedirs(day_folder)

    def test_02_make_folder_with_timestamp_as_input(self):
        """Make a new datastore folder in the base folder and that a DataFolder instance is returned."""
        # Arrange
        name = "temp"
        posix_time = time.time()
        tm = time.localtime(posix_time)
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        day_folder = os.path.join(os.getcwd(), date_str)
        expected_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
        try:
            # Act
            time.sleep(1)  # Sleep to change the current time timestamp
            datafolder = self.datastore.make_folder(name, posix_time)

            # Assert
            self.assertTrue(os.path.isdir(expected_folder))
            self.assertEqual(type(datafolder), DataFolder)

        finally:
            os.removedirs(expected_folder)

    def test_03_make_folder_with_date_and_time_as_inputs(self):
        """Make a new datastore folder in the base folder and that a DataFolder instance is returned."""
        # Arrange
        name = "temp"
        posix_time = time.time()
        tm = time.localtime(posix_time)
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        day_folder = os.path.join(os.getcwd(), date_str)
        expected_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
        try:
            # Act
            time.sleep(1)  # Sleep to change the current time timestamp
            datafolder = self.datastore.make_folder(name, date_str=date_str, time_str=time_str)

            # Assert
            self.assertTrue(os.path.isdir(expected_folder))
            self.assertEqual(type(datafolder), DataFolder)

        finally:
            os.removedirs(expected_folder)

    def test_04_make_folder_with_date_or_time_only_as_input_raises_exception(self):
        """See that making a new data folder ends with exception if both date and time are not present"""
        # Arrange
        name = "temp"
        posix_time = time.time()
        tm = time.localtime(posix_time)
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        # Act & Assert
        with self.assertRaises(ValueError):
            self.datastore.make_folder(name, date_str=date_str)

        with self.assertRaises(ValueError):
            self.datastore.make_folder(name, time_str=time_str)

    def test_05_make_folder_with_wrong_date_or_time_as_input_raises_exception(self):
        """See that making a new data folder ends with exception if both date and time are not present"""
        # Arrange
        name = "temp"
        posix_time = time.time()
        tm = time.localtime(posix_time)
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        bad_date_strings = [date_str[:-1], date_str + date_str[0], "1234567A"]
        bad_time_strings = [time_str[:-1], time_str + time_str[0], "65432I"]
        # Act & Assert
        for bad_date in bad_date_strings:
            with self.assertRaises(ValueError):
                self.datastore.make_folder(name, date_str=bad_date, time_str=time_str)

        for bad_time in bad_time_strings:
            with self.assertRaises(ValueError):
                self.datastore.make_folder(name, date_str=date_str, time_str=bad_time)

    def test_06_make_folder_with_missing_label_raises_exception(self):
        """See that making a new data folder ends with exception if proper label is not present"""
        # Arrange
        bad_names = ["", None, True, 0x04, "yö-öylätti"]
        posix_time = time.time()
        # Act & Assert
        for bad_name in bad_names:
            with self.assertRaises(ValueError):
                self.datastore.make_folder(bad_name, posix_time)

    def test_07_make_folder_raises_exception_when_directory_already_exists(self):
        """See that making the new data folder raises exception if `time` folder is already created"""
        name = "temp"
        posix_time = time.time()
        tm = time.localtime(posix_time)
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        day_folder = os.path.join(os.getcwd(), date_str)
        expected_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
        try:
            os.mkdir(day_folder)
            os.mkdir(expected_folder)
            # Act & Assert
            with self.assertRaises(FileExistsError):
                self.datastore.make_folder(name, date_str=date_str, time_str=time_str)

        finally:
            os.removedirs(expected_folder)

    def test_08_make_folder_only_name_as_input(self):
        """Make a new datastore folder in the base folder and that a DataFolder instance is returned."""
        # Arrange
        name = "temp"
        tm = time.localtime(time.time())
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        day_folder = os.path.join(os.getcwd(), date_str)
        expected_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
        try:
            # Act
            datafolder = self.datastore.make_folder(name)
            # We need to check for the rare case that the second just flipped between arranging the test and making
            # the folder. We ignore the super-rare case of this second being exactly at midnight, changing date also
            if not os.path.isdir(expected_folder):
                # Update to 'next' second
                tm = time.localtime(time.time())
                time_str = time.strftime("%H%M%S", tm)
                expected_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))

            # Assert
            self.assertTrue(os.path.isdir(expected_folder))
            self.assertEqual(type(datafolder), DataFolder)

        finally:
            os.removedirs(expected_folder)
            # os.removedirs(day_folder)

    def test_09_get_folder_with_name_date_and_time_as_inputs(self):
        """Get a datastore folder with the given input values."""
        # Arrange
        name = "temp"
        posix_time = time.time()
        tm = time.localtime(posix_time)
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        day_folder = os.path.join(os.getcwd(), date_str)
        expected_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
        self.datastore.make_folder(name, date_str=date_str, time_str=time_str)
        try:
            # Act
            datafolder = self.datastore.get_folder(name, date_str, time_str)
            # Assert
            self.assertEqual(type(datafolder), DataFolder)

        finally:
            os.removedirs(expected_folder)

    def test_10_get_folder_raises_exception_for_not_existing_folder(self):
        """See that exception is raised if a non-existing folder is tried to be found."""
        # Arrange
        name = "test"
        posix_time = time.time()
        tm = time.localtime(posix_time)
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        # Act & Assert
        with self.assertRaises(FileNotFoundError):
            self.datastore.get_folder(name, date_str, time_str)

    def test_11_get_folder_from_path(self):
        """Get a datastore folder with the given input values."""
        # Arrange
        name = "temp"
        posix_time = time.time()
        tm = time.localtime(posix_time)
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        day_folder = os.path.join(os.getcwd(), date_str)
        expected_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
        self.datastore.make_folder(name, date_str=date_str, time_str=time_str)
        try:
            # Act
            datafolder = self.datastore.get_folder_from_path(expected_folder)
            # Assert
            self.assertEqual(type(datafolder), DataFolder)

        finally:
            os.removedirs(expected_folder)

    def test_12_get_folder_from_path_raises_exception_for_not_existing_folder(self):
        """See that exception is raised if a non-existing folder is tried to be found."""
        # Arrange
        name = "test"
        posix_time = time.time()
        tm = time.localtime(posix_time)
        date_str = time.strftime("%Y%m%d", tm)
        time_str = time.strftime("%H%M%S", tm)
        day_folder = os.path.join(os.getcwd(), date_str)
        expected_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
        # Act & Assert
        with self.assertRaises(FileNotFoundError):
            self.datastore.get_folder_from_path(expected_folder)

    def test_13_list_folders(self):
        """Get a list of data folders created with specific label"""
        # Arrange
        name = "test"
        date_strings = ["20122012", "20022002", "20111102"]
        time_str = "000102"
        for date_str in date_strings:
            self.datastore.make_folder(name, date_str=date_str, time_str=time_str)

        # Act
        try:
            data_folders = self.datastore.list_folders(name)
            # Assert
            self.assertEqual(len(data_folders), len(date_strings))
            for data_folder in data_folders:
                self.assertEqual(type(data_folder), DataFolder)
                self.assertEqual(os.path.split(data_folder.folder_path)[-1], "{}_{}".format(time_str, name))
                self.assertIn(data_folder.date_str, date_strings)
                self.assertEqual(data_folder.label, name)
                self.assertEqual(data_folder.time_str, time_str)

        finally:
            for date_str in date_strings:
                day_folder = os.path.join(os.getcwd(), date_str)
                time_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
                os.removedirs(time_folder)

    def test_14_find_latest_folder(self):
        """Get a the latest data folder from created data folders with specific label"""
        # Arrange
        name = "test"
        date_strings = ["20122012", "20022002", "20111102"]
        time_strings = ["000102", "011234", "235959"]
        for date_str in date_strings:
            for time_str in time_strings:
                self.datastore.make_folder(name, date_str=date_str, time_str=time_str)

        latest_day = sorted(date_strings)[-1]
        latest_time = sorted(time_strings)[-1]

        # Act
        try:
            data_folder = self.datastore.find_latest_folder(name)
            # Assert
            self.assertEqual(type(data_folder), DataFolder)
            self.assertEqual(os.path.split(data_folder.folder_path)[-1], "{}_{}".format(latest_time, name))
            self.assertEqual(data_folder.date_str, latest_day)
            self.assertEqual(data_folder.label, name)
            self.assertEqual(data_folder.time_str, latest_time)

        finally:
            for date_str in date_strings:
                for time_str in time_strings:
                    day_folder = os.path.join(os.getcwd(), date_str)
                    time_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
                    os.removedirs(time_folder)

    def test_15_find_latest_with_specific_date(self):
        """Get a the latest data folder from created data folders with specific label and date"""
        # Arrange
        name = "test"
        date_strings = ["20122012", "20022002", "20111102"]
        time_strings = ["000102", "011234", "235959"]
        for date_str in date_strings:
            for time_str in time_strings:
                self.datastore.make_folder(name, date_str=date_str, time_str=time_str)

        expected_day = date_strings[-1]
        latest_time = sorted(time_strings)[-1]

        # Act
        try:
            data_folder = self.datastore.find_latest_folder(name, expected_day)
            # Assert
            self.assertEqual(type(data_folder), DataFolder)
            self.assertEqual(os.path.split(data_folder.folder_path)[-1], "{}_{}".format(latest_time, name))
            self.assertEqual(data_folder.date_str, expected_day)
            self.assertEqual(data_folder.label, name)
            self.assertEqual(data_folder.time_str, latest_time)

        finally:
            for date_str in date_strings:
                for time_str in time_strings:
                    day_folder = os.path.join(os.getcwd(), date_str)
                    time_folder = os.path.join(day_folder, "{}_{}".format(time_str, name))
                    os.removedirs(time_folder)


if __name__ == "__main__":
    unittest.main()



"""Instrument driver for the Imagine Eyes Mirao52e deformable mirror."""

import sys
import logging
import os
import re
import subprocess
from pathlib import Path

from typing import List, NamedTuple, Optional

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_UsageException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


# Named tuple representing mirror status.
MirrorStatus = NamedTuple("MirrorStatus", [
    ("psu_temp", float),
    ("locked", bool)
])


def read_mirao_file(filename: str) -> List[float]:
    """Read a set of mirror actuator values from an MRO file.

    The MRO files can be generated with the MroEdit tool.

    :param filename: File name of the MRO file.
    :return: List of 52 floating point values.
    """
    ret = 52 * [0.0]
    with open(filename, "r") as f:
        s = f.readline()
        s = s.strip()
        while (s.startswith("DLL_VERSION")
               or s.startswith("FILE_FORMAT_VERSION")
               or s.startswith("CREATION_DATE")
               or s.startswith("DISTRIBUTOR")):
            s = f.readline()
            s = s.strip()
        for i in range(52):
            m = re.match(r"^A([0-9]{2}):\s*([0-9a-fA-F]{4})\s*$", s)
            if not m:
                raise ValueError("Unexpected syntax in MRO file: {}".format(s))
            a = int(m.group(1))
            v = int(m.group(2), 16)
            if a != i:
                raise ValueError("Unexpected order of actuators in MRO file")
            if v > 0x3FFF:
                raise ValueError("Out-of-range actuator value in MRO file: 0x{:04x}".format(v))
            ret[i] = (v - 0x1FFF) / 8192.0
            s = f.readline()
            s = s.strip()
        if s != "END":
            raise ValueError("Unexpected syntax in MRO file: {}".format(s))
    return ret


class ImagineEyes_Mirao52e(QMI_Instrument):
    """Instrument driver for the Imagine Eyes Mirao52e deformable mirror.

    Note that the deformable mirror will only hold its actuator positions
    while the connection to the instrument is open. Closing the connection
    will automatically return the mirror to its free state (which is not flat).

    This driver communicates with the deformable mirror via
    a helper program "mirao_bridge.exe".

    The helper program is a command-line Windows application.
    On Linux, it can run under Wine.
    """

    # Location of bridge program.
    DEFAULT_BRIDGE_DIR = os.path.join(Path.home(), "mirao")
    BRIDGE_CMD = "mirao_bridge.exe"

    # USD vendor ID : product ID of the deformable mirror.
    USB_ID = "0403:6001"

    def __init__(self, context: QMI_Context, name: str, bridge_dir: str = DEFAULT_BRIDGE_DIR) -> None:
        """Initialize the driver.

        :param name: Name for this instrument instance.
        :param bridge_dir: Path to directory where mirao_bridge.exe is installed.
        """
        if not os.path.isdir(bridge_dir):
            raise QMI_InstrumentException("Mirao helper directory {!r} not found".format(bridge_dir))

        super().__init__(context, name)
        self._bridge_dir = bridge_dir
        self._proc = None  # type: Optional[subprocess.Popen]
        self._version_string = ""

    def _start_helper(self) -> None:
        """Start the helper program under native Windows."""
        assert self._proc is None
        self._proc = subprocess.Popen(args=self.BRIDGE_CMD,
                                      stdin=subprocess.PIPE,
                                      stdout=subprocess.PIPE,
                                      cwd=self._bridge_dir)

    def _start_helper_wine(self) -> None:
        """Start the helper program under wine."""
        assert self._proc is None
        env: dict = {}
        env.update(os.environ)
        env["WINEDLLPATH"] = "."
        env["FTDID"] = self.USB_ID
        self._proc = subprocess.Popen(args=["wine", self.BRIDGE_CMD],
                                      stdin=subprocess.PIPE,
                                      stdout=subprocess.PIPE,
                                      cwd=self._bridge_dir,
                                      env=env)
        assert self._proc.stdin is not None
        assert self._proc.stdout is not None

    def _stop_helper(self) -> None:
        """Stop the helper program."""
        assert self._proc is not None
        assert self._proc.stdin is not None
        assert self._proc.stdout is not None
        self._proc.stdin.close()
        self._proc.wait()
        self._proc.stdout.close()
        self._proc = None
        self._version_string = ""

    def _recv_response(self) -> List[str]:
        """Read response from helper program."""
        assert self._proc is not None
        assert self._proc.stdout is not None
        ret: List[str] = []
        while True:
            s = self._proc.stdout.readline()
            if not s:
                raise QMI_InstrumentException("Unexpected end of input from helper program")
            s = s.strip()
            _logger.debug("recv: %s", s)
            if s == b"READY":
                return ret
            ret.append(s.decode("latin1"))

    def _recv_handshake(self) -> None:
        """Receive handshake message from helper program."""
        lines = self._recv_response()
        if len(lines) == 1:
            w = lines[0].split()
            if len(w) == 2 and w[0] == "MIRAO52_BRIDGE":
                self._version_string = w[1]
                return
        raise QMI_InstrumentException("Unexpected handshake from helper program: {}".format(lines))

    def _send_cmd(self, cmd: str) -> None:
        """Send command to helper program and check response."""
        assert self._proc is not None
        assert self._proc.stdin is not None
        self._proc.stdin.write(cmd.encode("ascii") + b"\r\n")
        self._proc.stdin.flush()

    def _check_response(self) -> None:
        lines = self._recv_response()
        got_ok = False
        for s in lines:
            if s == "OK":
                got_ok = True
            if s.startswith("ERROR "):
                raise QMI_InstrumentException("Error from {}: {}".format(self._name, s[6:]))
        if not got_ok:
            raise QMI_InstrumentException("Unexpected response from {}: {}".format(self._name, lines))

    @rpc_method
    def open(self) -> None:

        self._check_is_closed()

        _logger.info("[%s] Starting helper program", self._name)

        if sys.platform.startswith("linux"):
            self._start_helper_wine()
        else:
            self._start_helper()

        try:
            self._recv_handshake()
            _logger.info("[%s] Opening connection to instrument", self._name)
            self._send_cmd("OPEN")
            self._check_response()
        except Exception:
            # Shut down helper program when an error occurs.
            self._stop_helper()
            raise

        super().open()

    @rpc_method
    def close(self) -> None:
        self._check_is_open()
        _logger.info("[%s] Closing connection to instrument", self._name)
        try:
            self._send_cmd("CLOSE")
            self._check_response()
        finally:
            _logger.info("[%s] Stopping helper program", self._name)
            self._stop_helper()
            super().close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        self._check_is_open()
        return QMI_InstrumentIdentification(vendor="ImagineEyes",
                                            model="Mirao52e",
                                            serial=None,
                                            version=self._version_string)

    @rpc_method
    def apply(self, actuators: List[float]) -> None:
        """Apply actuator values.

        :param actuators: List of 52 floating point values.
        """
        self._check_is_open()
        _logger.debug("[%s] Applying actuator values", self._name)
        if len(actuators) != 52:
            raise QMI_UsageException("Expecting 52 actuator values")
        cmd = "APPLY " + ",".join("{:.6f}".format(v) for v in actuators)
        self._send_cmd(cmd)
        self._check_response()

    @rpc_method
    def get_mirror_status(self) -> MirrorStatus:
        """Return the mirror status.

        :return: Named tuple (psu_temperature, locked)
            where psu_temperature is the PSU temperature in Celsius;
            locked is True if the mirror is locked due to an error condition.
        """
        self._check_is_open()
        self._send_cmd("MONITOR")
        lines = self._recv_response()
        psu_temp = 0.0
        locked = False
        got_resp = False
        for s in lines:
            m = re.match(r"^PSUTEMP=([0-9.]+) LOCKED=([01])$", s)
            if m:
                psu_temp = float(m.group(1))
                locked = (int(m.group(2)) != 0)
                got_resp = True
            if s.startswith("ERROR "):
                raise QMI_InstrumentException("Error from {}: {}".format(self._name, s[6:]))
        if not got_resp:
            raise QMI_InstrumentException("Unexpected response from {}: {}".format(self._name, lines))
        return MirrorStatus(psu_temp=psu_temp, locked=locked)



"""
Instrument driver for the four-channel Bristol Fiber-Optic Switch (FOS)
"""

import logging
import sys
import typing
from typing import Optional

import qmi.core.exceptions
from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method

# Lazy import of the "uldaq" or "ul" and "enums" modules. See the function _import_modules() below.
uldaq = None
ul, enums = None, None
if typing.TYPE_CHECKING:
    if sys.platform.startswith("linux") or sys.platform == "darwin":
        import uldaq  # type: ignore

    if sys.platform.startswith("win"):
        from mcculw import ul  # type: ignore
        from mcculw import enums  # type: ignore


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


def _import_modules() -> None:
    """Import the "uldaq" library or "mcculw" modules.

    This import is done in a function, instead of at the top-level, to avoid unnecessary
    dependencies for programs that do not access the instrument directly.
    """
    global uldaq, ul, enums
    _logger.debug("Importing %s modules", sys.platform)
    if (sys.platform.startswith("linux") or sys.platform == "darwin") and uldaq is None:
        import uldaq  # type: ignore

    elif sys.platform.startswith("win") and (ul is None or enums is None):
        from mcculw import ul, enums  # type: ignore


class _Bristol_FosUnix:
    """Unix version for the Bristol FOS instrument driver."""
    def __init__(self, unique_id: str) -> None:
        self._unique_id = unique_id
        self._device = None
        self._dio_device = None

    @property
    def device(self) -> "uldaq.DaqDevice":  # type: ignore
        """Property for holding the DAQ device object."""
        assert self._device is not None
        return self._device

    @property
    def dio_device(self) -> "uldaq.DioDevice":  # type: ignore
        """Property for holding the DIO device object."""
        assert self._dio_device is not None
        return self._dio_device

    @staticmethod
    def _find_device_descriptor(unique_id: str) -> "Optional[uldaq.DaqDeviceDescriptor]":  # type: ignore
        """A method to retrieve a specific instrument's 'device descriptor' object based on unique ID of the instrument.

        Parameters:
            unique_id:         A unique ID string.

        Returns:
            device_descriptor: The device descriptor with matching unique ID or None.
        """
        assert uldaq is not None
        device_descriptors = uldaq.get_daq_device_inventory(uldaq.InterfaceType.ANY)
        for device_descriptor in device_descriptors:
            if device_descriptor.unique_id == unique_id:
                return device_descriptor

        return None  # Device not found.

    @rpc_method
    def open(self) -> None:
        device_descriptor = self._find_device_descriptor(self._unique_id)
        if device_descriptor is None:
            _logger.error("Bristol FOS with unique_id '%s' not found.", self._unique_id)
            raise ValueError(f"Bristol FOS with unique_id '{self._unique_id!r}' not found.")

        assert uldaq is not None
        device = uldaq.DaqDevice(device_descriptor)
        try:
            device.connect()
            try:
                dio_device = device.get_dio_device()
                dio_device.d_config_port(uldaq.DigitalPortType.FIRSTPORTA, uldaq.DigitalDirection.OUTPUT)

            except Exception as exc:
                device.disconnect()
                raise Exception from exc

        except Exception as exc:
            device.release()
            _logger.error("Bristol FOS device connection failed with: %s", str(exc))
            raise qmi.core.exceptions.QMI_InstrumentException("Bristol FOS device connection failed.") from exc

        self._device = device
        self._dio_device = dio_device

    @rpc_method
    def close(self) -> None:
        self.device.disconnect()
        self.device.release()
        self._device = None
        self._dio_device = None

    @rpc_method
    def select_channel(self, channel: int) -> None:
        assert uldaq is not None
        # Note that the 'channel parameter has values 1..4 ; these are mapped to value 0..3 here.
        self.dio_device.d_out(uldaq.DigitalPortType.FIRSTPORTA, channel - 1)


class _Bristol_FosWindows:
    """Windows version for the Bristol FOS instrument driver"""

    def __init__(self, unique_id: str, board_id: int) -> None:
        self._unique_id = unique_id
        self._board_id = board_id

    @property
    def board_id(self) -> int:
        """A property for instrument board number. It must be within [0, 99]."""
        assert 0 <= self._board_id <= 99
        return self._board_id

    @board_id.setter
    def board_id(self, new_id: int) -> None:
        """A property for setting a new instrument board number. It must be within [0, 99]."""
        assert isinstance(new_id, int)
        assert 0 <= new_id <= 99
        self._board_id = new_id

    @staticmethod
    def _find_device_descriptor(unique_id: str) -> "Optional[structs.DaqDeviceDescriptor]":  # type: ignore
        """A method to retrieve a specific instrument's 'device descriptor' object based on unique ID of the instrument.

        Parameters:
            unique_id:         A unique ID string.

        Returns:
            device_descriptor: The device descriptor with matching unique ID or None.
        """
        assert ul is not None and enums is not None
        device_descriptors = ul.get_daq_device_inventory(enums.InterfaceType.ANY)
        for device_descriptor in device_descriptors:
            if device_descriptor.unique_id == unique_id:
                return device_descriptor

        return None  # Device not found.

    @rpc_method
    def open(self) -> None:
        assert ul is not None and enums is not None
        ul.ignore_instacal()  # With this we ignore 'cg.cfg' file and enable runtime configuring.
        device_descriptor = self._find_device_descriptor(self._unique_id)
        if device_descriptor is None:
            _logger.error("Bristol FOS with unique_id '%s' not found.", self._unique_id)
            raise ValueError(f"Bristol FOS with unique_id '{self._unique_id!r}' not found.")

        try:
            ul.create_daq_device(self.board_id, device_descriptor)
            assert self.board_id == ul.get_board_number(
                device_descriptor
            ), f"{self.board_id} != {ul.get_board_number(device_descriptor)}"
            ul.d_config_port(self.board_id, enums.DigitalPortType.FIRSTPORTA, enums.DigitalIODirection.OUT)

        except Exception as exc:
            _logger.error("Bristol FOS device configuration failed with: %s", str(exc))
            ul.release_daq_device(self.board_id)
            raise qmi.core.exceptions.QMI_InstrumentException("Bristol FOS device port configuration failed.") from exc

    @rpc_method
    def close(self) -> None:
        assert ul is not None
        ul.release_daq_device(self.board_id)

    @rpc_method
    def select_channel(self, channel: int) -> None:
        assert ul is not None
        # Note that the 'channel parameter has values 1..4 ; these are mapped to value 0..3 here.
        ul.d_out(self._board_id, enums.DigitalPortType.FIRSTPORTA, channel - 1)


class Bristol_Fos(QMI_Instrument):
    """Base class for the Bristol FOS instrument driver"""

    def __init__(self, context: QMI_Context, name: str, unique_id: str, board_id: int = 0) -> None:
        """Base class initialization. Depending on the system platform, a Windows-compatible or
        Unix-compatible driver version is instantiated.

        Attributes:
            fos:        The instantiated FOS device on the driver.

        Parameters:
            context:    A QMI_Context instance.
            name:       Name for the instrument in the context.
            unique_id:  An unique identification number, a.k.a. serial number of the device.
            board_id:   Board number for Windows driver. Not used for Unix driver.
        """
        super().__init__(context, name)
        if sys.platform.startswith("win"):
            self.fos = _Bristol_FosWindows(unique_id, board_id)

        else:
            self.fos = _Bristol_FosUnix(unique_id)

        # Import the support module.
        _import_modules()

    @rpc_method
    def open(self) -> None:
        self._check_is_closed()
        self.fos.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        self._check_is_open()
        self.fos.close()
        super().close()

    @rpc_method
    def select_channel(self, channel: int) -> None:
        """Method for selecting an output channel for DIO device.

        Parameters:
            channel:     The output channel number. Must be in range [1, 4].

        Raises:
            ValueError:  At invalid channel number.
        """
        self._check_is_open()
        if channel not in range(1, 5):
            raise ValueError("Bad channel: {}".format(channel))

        self.fos.select_channel(channel)



#! /usr/bin/env python3

"""Test publish/subscribe functionality."""
import random
import time
import unittest

import qmi
import qmi.core.exceptions
from qmi.core.config_defs import CfgQmi, CfgContext
from qmi.core.context import QMI_Context
from qmi.core.rpc import QMI_RpcObject, rpc_method
from qmi.core.pubsub import SignalDescription, QMI_Signal, QMI_SignalReceiver


class MyPublisher(QMI_RpcObject):

    last_instance = None

    sig1 = QMI_Signal([])
    sig2 = QMI_Signal([int])
    sig3 = QMI_Signal([int, str])

    def __init__(self, context, name):
        super().__init__(context, name)
        MyPublisher.last_instance = self

    @rpc_method
    def send1(self):
        self.sig1.publish()

    @rpc_method
    def send2(self):
        self.sig2.publish(24)

    @rpc_method
    def send3(self, x, s):
        self.sig3.publish(x, s)

    @rpc_method
    def send_slow(self):
        for i in range(5):
            time.sleep(0.5)
            self.sig2.publish(i)


class TestLocalPubSub(unittest.TestCase):
    """Test publish/subscribe within the local context."""

    def setUp(self):
        self.context_name = "test_local_pubsub"
        qmi.start(self.context_name)

    def tearDown(self):
        qmi.stop()

    def test_register_signal(self):
        # Test registration of signals.

        pub1 = qmi.make_rpc_object("pub1", MyPublisher)
        pub2 = qmi.make_rpc_object("pub2", MyPublisher)

        # Check list of registered signals.
        sigs = pub1.get_signals()
        expect_sigs = {
            SignalDescription("sig1", ()),
            SignalDescription("sig2", (int,)),
            SignalDescription("sig3", (int, str))
        }
        self.assertEqual(set(sigs), expect_sigs)

        sigs = pub2.get_signals()
        self.assertEqual(set(sigs), expect_sigs)

    def test_subscribe(self):
        # Test subscribing to signals.

        qmi.make_rpc_object("pub1", MyPublisher)
        qmi.make_rpc_object("pub2", MyPublisher)

        # Create signal receiver queue.
        recv = QMI_SignalReceiver()
        recv2 = QMI_SignalReceiver()

        # Subscribe to signals.
        qmi.context().subscribe_signal(self.context_name, "pub1", "sig1", recv)
        qmi.context().subscribe_signal(self.context_name, "pub2", "sig2", recv)
        qmi.context().subscribe_signal(self.context_name, "pub1", "sig1", recv2)

        # Unsubscribe from signals.
        qmi.context().unsubscribe_signal(self.context_name, "pub1", "sig1", recv)
        qmi.context().unsubscribe_signal(self.context_name, "pub2", "sig2", recv)
        qmi.context().unsubscribe_signal(self.context_name, "pub1", "sig1", recv2)

        # Check that receiver queues are still empty.
        self.assertFalse(recv.has_signal_ready())
        self.assertFalse(recv2.has_signal_ready())

        # Try to unsubscribe from a non-subscribed signal - should be ignored.
        qmi.context().unsubscribe_signal(self.context_name, "pub1", "sig1", recv)

    def test_pubsub(self):
        # Publish and receive signals.

        qmi.make_rpc_object("pub1", MyPublisher)
        qmi.make_rpc_object("pub2", MyPublisher)

        # Create signal receiver queue.
        recv = QMI_SignalReceiver()

        # Subscribe to signals.
        qmi.context().subscribe_signal(self.context_name, "pub1", "sig1", recv)
        qmi.context().subscribe_signal(self.context_name, "pub1", "sig2", recv)
        qmi.context().subscribe_signal(self.context_name, "pub2", "sig1", recv)

        # Publish signals.
        qmi.context().publish_signal("pub1", "sig1", 111)
        qmi.context().publish_signal("pub1", "sig1", 112)
        qmi.context().publish_signal("pub1", "sig2", "aap", "noot")

        # Check that signals are received.
        self.assertTrue(recv.has_signal_ready())
        self.assertEqual(recv.get_queue_length(), 3)
        sig = recv.get_next_signal()
        self.assertEqual(sig.publisher_context, self.context_name)
        self.assertEqual(sig.publisher_name, "pub1")
        self.assertEqual(sig.signal_name, "sig1")
        self.assertEqual(sig.args, (111,))
        self.assertEqual(sig.receiver_seqnr, 0)
        sig = recv.get_next_signal()
        self.assertEqual(sig.publisher_context, self.context_name)
        self.assertEqual(sig.publisher_name, "pub1")
        self.assertEqual(sig.signal_name, "sig1")
        self.assertEqual(sig.args, (112,))
        self.assertEqual(sig.receiver_seqnr, 1)
        sig = recv.get_next_signal()
        self.assertEqual(sig.publisher_context, self.context_name)
        self.assertEqual(sig.publisher_name, "pub1")
        self.assertEqual(sig.signal_name, "sig2")
        self.assertEqual(sig.args, ("aap", "noot"))
        self.assertEqual(sig.receiver_seqnr, 2)

        # Check no more signals in queue.
        self.assertFalse(recv.has_signal_ready())
        self.assertEqual(recv.get_queue_length(), 0)

        # Try to get signal from queue - should time out.
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            sig = recv.get_next_signal(timeout=1.0)

        # Publish a non-subscribed signal - should be fine.
        qmi.context().publish_signal("pub2", "sig2", 211)

        # Check no new signals received.
        self.assertFalse(recv.has_signal_ready())

        # Unsubscribe from one of the signals.
        qmi.context().unsubscribe_signal(self.context_name, "pub1", "sig1", recv)

        # Publish more signals.
        qmi.context().publish_signal("pub1", "sig1", 113)
        qmi.context().publish_signal("pub1", "sig2", "mies", "wim")

        # Check received as expected.
        self.assertTrue(recv.has_signal_ready())
        self.assertEqual(recv.get_queue_length(), 1)
        sig = recv.get_next_signal()
        self.assertEqual(sig.publisher_context, self.context_name)
        self.assertEqual(sig.publisher_name, "pub1")
        self.assertEqual(sig.signal_name, "sig2")
        self.assertEqual(sig.args, ("mies", "wim"))
        self.assertEqual(sig.receiver_seqnr, 3)

        # Unsubscribe from all signals.
        qmi.context().unsubscribe_signal(self.context_name, "pub1", "sig2", recv)
        qmi.context().unsubscribe_signal(self.context_name, "pub2", "sig1", recv)

        # Publish more signals.
        qmi.context().publish_signal("pub1", "sig1", 114)
        qmi.context().publish_signal("pub1", "sig2", "zus", "jet")
        qmi.context().publish_signal("pub2", "sig1", 3.14)

        # Check nothing received.
        self.assertFalse(recv.has_signal_ready())

    def test_subscribe_proxy(self):
        # Test subscribing to signals via RPC proxy.

        pub1 = qmi.make_rpc_object("pub1", MyPublisher)

        # Create signal receiver queue.
        recv = QMI_SignalReceiver()
        recv2 = QMI_SignalReceiver()

        # Subscribe to signals (registered and unregistered).
        pub1.sig1.subscribe(recv)
        pub1.sig2.subscribe(recv)
        pub1.sig1.subscribe(recv2)

        # Unsubscribe from signals.
        pub1.sig1.unsubscribe(recv)
        pub1.sig2.unsubscribe(recv)
        pub1.sig1.unsubscribe(recv2)

        # Check that receiver queues are still empty.
        self.assertFalse(recv.has_signal_ready())
        self.assertFalse(recv2.has_signal_ready())

        # Try to unsubscribe from a non-subscribed signal - should be ignored.
        pub1.sig2.unsubscribe(recv2)

    def test_pubsub_rpc(self):
        # Publish signals via RPC trigger and receive signals.

        pub1 = qmi.make_rpc_object("pub1", MyPublisher)
        pub2 = qmi.make_rpc_object("pub2", MyPublisher)

        # Create signal receiver queue.
        recv = QMI_SignalReceiver()

        # Subscribe to signals.
        pub1.sig1.subscribe(recv)
        pub1.sig2.subscribe(recv)
        pub2.sig1.subscribe(recv)

        # Publish signals.
        pub1.send1()
        pub1.send1()
        pub1.send2()
        pub2.send1()

        # Check that signals are received.
        self.assertTrue(recv.has_signal_ready())
        self.assertEqual(recv.get_queue_length(), 4)
        sig = recv.get_next_signal()
        self.assertEqual(sig.publisher_context, self.context_name)
        self.assertEqual(sig.publisher_name, "pub1")
        self.assertEqual(sig.signal_name, "sig1")
        self.assertEqual(sig.args, ())
        self.assertEqual(sig.receiver_seqnr, 0)
        sig = recv.get_next_signal()
        self.assertEqual(sig.publisher_context, self.context_name)
        self.assertEqual(sig.publisher_name, "pub1")
        self.assertEqual(sig.signal_name, "sig1")
        self.assertEqual(sig.args, ())
        self.assertEqual(sig.receiver_seqnr, 1)
        sig = recv.get_next_signal()
        self.assertEqual(sig.publisher_context, self.context_name)
        self.assertEqual(sig.publisher_name, "pub1")
        self.assertEqual(sig.signal_name, "sig2")
        self.assertEqual(sig.args, (24,))
        self.assertEqual(sig.receiver_seqnr, 2)
        sig = recv.get_next_signal()
        self.assertEqual(sig.publisher_context, self.context_name)
        self.assertEqual(sig.publisher_name, "pub2")
        self.assertEqual(sig.signal_name, "sig1")
        self.assertEqual(sig.args, ())
        self.assertEqual(sig.receiver_seqnr, 3)

        # Check no more signals in queue.
        self.assertFalse(recv.has_signal_ready())
        self.assertEqual(recv.get_queue_length(), 0)

    def test_multi_sub(self):
        # Multiple subscribers per signal.

        pub1 = qmi.make_rpc_object("pub1", MyPublisher)
        pub2 = qmi.make_rpc_object("pub2", MyPublisher)

        # Create signal receiver queues.
        recv1 = QMI_SignalReceiver()
        recv2 = QMI_SignalReceiver()
        recv3 = QMI_SignalReceiver()

        # Subscribe to signals.
        pub1.sig2.subscribe(recv1)
        pub1.sig3.subscribe(recv1)
        pub1.sig3.subscribe(recv2)
        pub2.sig2.subscribe(recv3)
        pub2.sig3.subscribe(recv1)
        pub2.sig3.subscribe(recv2)
        pub2.sig3.subscribe(recv3)

        # Publish signals.
        pub1.send2()
        pub1.send3(21, "one")
        pub2.send2()
        pub2.send3(41, "two")
        pub1.send2()
        pub1.send3(22, "")
        pub2.send2()
        pub2.send3(42, "")

        # Check that signals are received as expected.
        self.assertTrue(recv1.has_signal_ready())
        self.assertEqual(recv1.get_queue_length(), 6)
        sig = recv1.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig2", (24,), 0))
        sig = recv1.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig3", (21, "one"), 1))
        sig = recv1.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub2", "sig3", (41, "two"), 2))
        sig = recv1.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig2", (24,), 3))
        sig = recv1.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig3", (22, ""), 4))
        sig = recv1.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub2", "sig3", (42, ""), 5))

        self.assertTrue(recv2.has_signal_ready())
        self.assertEqual(recv2.get_queue_length(), 4)
        sig = recv2.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig3", (21, "one"), 0))
        sig = recv2.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub2", "sig3", (41, "two"), 1))
        sig = recv2.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig3", (22, ""), 2))
        sig = recv2.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub2", "sig3", (42, ""), 3))

        self.assertTrue(recv3.has_signal_ready())
        self.assertEqual(recv3.get_queue_length(), 4)
        sig = recv3.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub2", "sig2", (24,), 0))
        sig = recv3.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub2", "sig3", (41, "two"), 1))
        sig = recv3.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub2", "sig2", (24,), 2))
        sig = recv3.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub2", "sig3", (42, ""), 3))

        # Unsubscribe some receivers.
        pub1.sig3.unsubscribe(recv2)
        pub2.sig2.unsubscribe(recv3)
        pub2.sig3.unsubscribe(recv2)
        pub2.sig3.unsubscribe(recv3)

        # Publish more signals.
        qmi.context().publish_signal("pub2", "sig3", 43, "three")

        # Check that signals are received as expected.
        self.assertTrue(recv1.has_signal_ready())
        self.assertEqual(recv1.get_queue_length(), 1)
        sig = recv1.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub2", "sig3", (43, "three"), 6))

        # Check no further signals received.
        self.assertFalse(recv1.has_signal_ready())
        self.assertFalse(recv2.has_signal_ready())
        self.assertFalse(recv3.has_signal_ready())

    def test_queue_overflow(self):
        # Test receiver queue overflow behaviour.

        pub1 = qmi.make_rpc_object("pub1", MyPublisher)

        # Create signal receiver queue.
        recv = QMI_SignalReceiver(max_queue_length=5, discard_policy=QMI_SignalReceiver.DISCARD_OLD)

        # Subscribe to signal.
        pub1.sig3.subscribe(recv)

        # Publish signals.
        for i in range(10):
            pub1.send3(10 + i, "")

        # Check that only the last 5 signals remain in the queue.
        self.assertEqual(recv.get_queue_length(), 5)
        sig = recv.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig3", (15, ""), 5))

        # Publish another signal - should fit in the queue without overflow.
        pub1.send3(20, "")

        # Check.
        self.assertEqual(recv.get_queue_length(), 5)
        sig = recv.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig3", (16, ""), 6))

        # Publish more signals - should push old signals out of the queue.
        for i in range(3):
            pub1.send3(30 + i, "")

        # Check queue contents.
        self.assertEqual(recv.get_queue_length(), 5)
        sig = recv.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig3", (19, ""), 9))
        sig = recv.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig3", (20, ""), 10))
        for i in range(3):
            sig = recv.get_next_signal()
            self.assertEqual(sig, (self.context_name, "pub1", "sig3", (30 + i, ""), 11 + i))

        self.assertFalse(recv.has_signal_ready())
        self.assertEqual(recv.get_queue_length(), 0)

        pub1.sig3.unsubscribe(recv)

        # Try again, this time with DISCARD_NEW policy.
        recv = QMI_SignalReceiver(max_queue_length=5, discard_policy=QMI_SignalReceiver.DISCARD_NEW)
        pub1.sig3.subscribe(recv)

        # Publish signals.
        for i in range(10):
            pub1.send3(10 + i, "")

        # Check that only the first 5 signals remain in the queue.
        self.assertEqual(recv.get_queue_length(), 5)
        sig = recv.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig3", (10, ""), 0))

        # Publish more signals - only one of them can fit in the queue.
        for i in range(3):
            pub1.send3(30 + i, "")

        # Check queue contents.
        self.assertEqual(recv.get_queue_length(), 5)
        for i in range(4):
            sig = recv.get_next_signal()
            self.assertEqual(sig, (self.context_name, "pub1", "sig3", (11 + i, ""), 1 + i))
        sig = recv.get_next_signal()
        self.assertEqual(sig, (self.context_name, "pub1", "sig3", (30, ""), 10))

        self.assertFalse(recv.has_signal_ready())
        self.assertEqual(recv.get_queue_length(), 0)

    def test_async(self):
        # Test asynchronous delivery of signals.

        pub1 = qmi.make_rpc_object("pub1", MyPublisher)

        # Create receiver queue.
        recv = QMI_SignalReceiver()

        # Subscribe to signal.
        pub1.sig2.subscribe(recv)

        # Start a slow RPC command which publishes signals from the RPC thread.
        future = pub1.rpc_nonblocking.send_slow()

        # Check no signals received yet.
        self.assertFalse(recv.has_signal_ready())

        # Blocking wait for first signal.
        sig = recv.get_next_signal(timeout=2.0)
        self.assertEqual(sig, (self.context_name, "pub1", "sig2", (0,), 0))

        # Check no further signals received.
        self.assertFalse(recv.has_signal_ready())

        # Sleep until second signal received.
        time.sleep(0.75)
        self.assertEqual(recv.get_queue_length(), 1)

        # Check second signal.
        sig = recv.get_next_signal(timeout=0)
        self.assertEqual(sig, (self.context_name, "pub1", "sig2", (1,), 1))

        # Blocking wait for 3rd signal.
        sig = recv.get_next_signal(timeout=2.0)
        self.assertEqual(sig, (self.context_name, "pub1", "sig2", (2,), 2))

        # Check no further signals ready.
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            recv.get_next_signal(timeout=0.1)

        # Blocking wait for 4th signal.
        sig = recv.get_next_signal(timeout=2.0)
        self.assertEqual(sig, (self.context_name, "pub1", "sig2", (3,), 3))

        # Finish the RPC call.
        future.wait()

        # Check 5th signal.
        sig = recv.get_next_signal(timeout=0)
        self.assertEqual(sig, (self.context_name, "pub1", "sig2", (4,), 4))

    def test_subscribe_unknown_object(self):
        # Subscribing to a non-existing object gives an error.

        pub1 = qmi.make_rpc_object("pub1", MyPublisher)
        recv = QMI_SignalReceiver()

        qmi.context().remove_rpc_object(pub1)

        with self.assertRaises(qmi.core.exceptions.QMI_SignalSubscriptionException):
            pub1.sig1.subscribe(recv)

    def test_remove_publisher(self):
        # Remove a publisher while a receiver is subscribed to its signal.

        # Create object and subscribe to its signal.
        pub1 = qmi.make_rpc_object("pub1", MyPublisher)

        recv = QMI_SignalReceiver()
        pub1.sig1.subscribe(recv)

        # Publish a signal and check that it is received.
        pub1.send1()
        sig = recv.get_next_signal(timeout=0)
        self.assertEqual(sig, (self.context_name, "pub1", "sig1", (), 0))

        # Remove publisher.
        qmi.context().remove_rpc_object(pub1)

        # Create new object with the same name.
        pub1 = qmi.make_rpc_object("pub1", MyPublisher)

        # Publish a signal and check that it is not received.
        pub1.send1()
        self.assertFalse(recv.has_signal_ready())

        # Re-subscribe to the signal.
        pub1.sig1.subscribe(recv)

        # Publish a signal and check that it is received.
        pub1.send1()
        sig = recv.get_next_signal(timeout=0)
        self.assertEqual(sig, (self.context_name, "pub1", "sig1", (), 1))


class TestRemotePubSub(unittest.TestCase):
    """Test publish/subscribe between contexts."""

    def setUp(self):

        # Initialize two QMI contexts.
        config = CfgQmi(
            contexts={
                "context1": CfgContext(tcp_server_port=0),
                "context2": CfgContext()
            }
        )
        self.context1 = QMI_Context("context1", config)
        self.context2 = QMI_Context("context2", config)
        self.context1.start()
        self.context2.start()

        # Context2 connects to context1.
        context1_address = "localhost:{}".format(self.context1.get_tcp_server_port())
        self.context2.connect_to_peer("context1", context1_address)

    def tearDown(self):
        self.context1.stop()
        self.context2.stop()

    def test_pubsub_remote(self):
        # Publish and receive signals between contexts.

        # Create publisher in context1.
        pub1 = self.context1.make_rpc_object("pub1", MyPublisher)

        # Create publisher in context2.
        pub2 = self.context2.make_rpc_object("pub2", MyPublisher)

        # Check that context2 can see the signals in context1.
        proxy = self.context2.get_rpc_object_by_name("context1.pub1")
        sigs = proxy.get_signals()
        expect_sigs = {
            SignalDescription("sig1", ()),
            SignalDescription("sig2", (int,)),
            SignalDescription("sig3", (int, str))
        }
        self.assertEqual(set(sigs), expect_sigs)

        # Create signal receiver queue.
        recv1 = QMI_SignalReceiver()
        recv2 = QMI_SignalReceiver()

        # Subscribe to signals.
        # NOTE: context1 can not subscribe to signals from context2
        #       because context1 has not connected to context2.
        pub1.sig3.subscribe(recv1)
        pub2.sig3.subscribe(recv2)
        proxy.sig3.subscribe(recv2)

        # Publish signals.
        pub2.send3(21, "A")
        pub1.send3(11, "B")
        proxy.send2()

        # Wait until remote signals delivered.
        time.sleep(0.2)

        # Check that signals are received.
        self.assertTrue(recv1.has_signal_ready())
        self.assertEqual(recv1.get_queue_length(), 1)
        sig = recv1.get_next_signal()
        self.assertEqual(sig, ("context1", "pub1", "sig3", (11, "B"), 0))

        self.assertTrue(recv2.has_signal_ready())
        self.assertEqual(recv2.get_queue_length(), 2)
        sig = recv2.get_next_signal()
        self.assertEqual(sig, ("context2", "pub2", "sig3", (21, "A"), 0))
        sig = recv2.get_next_signal()
        self.assertEqual(sig, ("context1", "pub1", "sig3", (11, "B"), 1))

        # Unsubscribe from remote signal.
        proxy.sig3.unsubscribe(recv2)

        # Publish signals.
        pub1.send3(12, "C")
        pub2.send3(22, "D")

        # Wait until remote signals delivered (if any).
        time.sleep(0.2)

        # Check that local signals are received.
        self.assertTrue(recv1.has_signal_ready())
        self.assertEqual(recv1.get_queue_length(), 1)
        sig = recv1.get_next_signal()
        self.assertEqual(sig, ("context1", "pub1", "sig3", (12, "C"), 1))

        self.assertTrue(recv2.has_signal_ready())
        self.assertEqual(recv2.get_queue_length(), 1)
        sig = recv2.get_next_signal()
        self.assertEqual(sig, ("context2", "pub2", "sig3", (22, "D"), 2))

    def test_remote_async(self):
        # Test asynchronous delivery of signals between contexts.

        # Create publisher in context1 with proxy in context2.
        pub1 = self.context1.make_rpc_object("pub1", MyPublisher)
        proxy = self.context2.get_rpc_object_by_name("context1.pub1")

        # Create receiver queue.
        recv = QMI_SignalReceiver()

        # Subscribe to signal.
        proxy.sig2.subscribe(recv)

        # Start a slow RPC command which publishes signals from the RPC thread.
        future = pub1.rpc_nonblocking.send_slow()

        # Check no signals received yet.
        self.assertFalse(recv.has_signal_ready())

        # Blocking wait for first signal.
        sig = recv.get_next_signal(timeout=2.0)
        self.assertEqual(sig, ("context1", "pub1", "sig2", (0,), 0))

        # Check no further signals received.
        self.assertFalse(recv.has_signal_ready())

        # Sleep until second signal received.
        time.sleep(0.75)
        self.assertEqual(recv.get_queue_length(), 1)

        # Check second signal.
        sig = recv.get_next_signal(timeout=0)
        self.assertEqual(sig, ("context1", "pub1", "sig2", (1,), 1))

        # Blocking wait for 3rd signal.
        sig = recv.get_next_signal(timeout=2.0)
        self.assertEqual(sig, ("context1", "pub1", "sig2", (2,), 2))

        # Check no further signals ready.
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            sig = recv.get_next_signal(timeout=0.1)

        # Blocking wait for 4th signal.
        sig = recv.get_next_signal(timeout=2.0)
        self.assertEqual(sig, ("context1", "pub1", "sig2", (3,), 3))

        # Finish the RPC call.
        future.wait()

        # Check 5th signal.
        sig = recv.get_next_signal(timeout=0.2)
        self.assertEqual(sig, ("context1", "pub1", "sig2", (4,), 4))

    def test_subscribe_unknown_remote_object(self):
        # Subscribing to a non-existing object gives an error.

        # Create object in context1 with proxy in context2.
        pub1 = self.context1.make_rpc_object("pub1", MyPublisher)
        proxy = self.context2.get_rpc_object_by_name("context1.pub1")

        # Remove the object in context1.
        self.context1.remove_rpc_object(pub1)

        # Try to subscribe in context2 to a signal from the removed object.
        recv = QMI_SignalReceiver()
        with self.assertRaises(qmi.core.exceptions.QMI_SignalSubscriptionException):
            proxy.sig1.subscribe(recv)

    def test_remove_remote_publisher(self):
        # Remove a publisher while a receiver is subscribed to its signal.

        # Create object in context1 with proxy in context2.
        pub1 = self.context1.make_rpc_object("pub1", MyPublisher)
        proxy = self.context2.get_rpc_object_by_name("context1.pub1")

        # Subscribe in context2 to a signal published in context1.
        recv = QMI_SignalReceiver()
        proxy.sig1.subscribe(recv)

        # Publish a signal and check that it is received.
        pub1.send1()
        sig = recv.get_next_signal(timeout=0.2)
        self.assertEqual(sig, ("context1", "pub1", "sig1", (), 0))

        # Remove publisher.
        self.context1.remove_rpc_object(pub1)

        # Create new object with the same name.
        pub1 = self.context1.make_rpc_object("pub1", MyPublisher)

        # Publish a signal and check that it is not received.
        pub1.send1()
        time.sleep(0.2)
        self.assertFalse(recv.has_signal_ready())

        # Re-subscribe to the signal.
        proxy = self.context2.get_rpc_object_by_name("context1.pub1")
        proxy.sig1.subscribe(recv)

        # Publish a signal and check that it is received.
        pub1.send1()
        sig = recv.get_next_signal(timeout=0.2)
        self.assertEqual(sig, ("context1", "pub1", "sig1", (), 1))

    def test_disconnect_remote_publisher(self):
        # Subscribing to a signal in a disconnected context gives an error.

        # Create object in context1 with proxy in context2.
        pub1 = self.context1.make_rpc_object("pub1", MyPublisher)
        proxy = self.context2.get_rpc_object_by_name("context1.pub1")

        # Subscribe in context2 to a signal published in context1.
        recv = QMI_SignalReceiver()
        proxy.sig1.subscribe(recv)

        # Publish a signal and check that it is received.
        pub1.send1()
        sig = recv.get_next_signal(timeout=0.2)
        self.assertEqual(sig, ("context1", "pub1", "sig1", (), 0))

        # Close the connection between the contexts.
        self.context2.disconnect_from_peer("context1")
        time.sleep(0.2)

        # Publish a signal and check that it is not received.
        pub1.send1()
        time.sleep(0.2)
        self.assertFalse(recv.has_signal_ready())

        # Try to resubscribe to the signal in the disconnected context.
        with self.assertRaises(qmi.core.exceptions.QMI_SignalSubscriptionException):
            proxy.sig1.subscribe(recv)
        with self.assertRaises(qmi.core.exceptions.QMI_SignalSubscriptionException):
            proxy.sig2.subscribe(recv)

    def test_multiple_remote_subscriptions(self):
        # Test multiple subscriptions to the same remote signal.

        # Create object in context1 with proxy in context2.
        pub1 = self.context1.make_rpc_object("pub1", MyPublisher)
        proxy = self.context2.get_rpc_object_by_name("context1.pub1")

        # Subscribe two receivers in context2 to the same signal from context1.
        recv1 = QMI_SignalReceiver()
        recv2 = QMI_SignalReceiver()
        proxy.sig3.subscribe(recv1)
        proxy.sig3.subscribe(recv2)

        # Publish signal, check that both receivers get it.
        pub1.send3(11, "A")
        sig = recv1.get_next_signal(timeout=0.2)
        self.assertEqual(sig, ("context1", "pub1", "sig3", (11, "A"), 0))
        sig = recv2.get_next_signal(timeout=0.2)
        self.assertEqual(sig, ("context1", "pub1", "sig3", (11, "A"), 0))

        # Unsubscribe one receiver.
        proxy.sig3.unsubscribe(recv2)

        # Check that the other receiver still gets signals.
        time.sleep(0.2)
        pub1.send3(12, "B")
        sig = recv1.get_next_signal(timeout=0.2)
        self.assertEqual(sig, ("context1", "pub1", "sig3", (12, "B"), 1))
        self.assertFalse(recv2.has_signal_ready())

        # Resubscribe, then unsubscribe the other receiver.
        proxy.sig3.subscribe(recv2)
        proxy.sig3.unsubscribe(recv1)

        # Check that signals are still correctly delivered.
        time.sleep(0.2)
        pub1.send3(13, "C")
        sig = recv2.get_next_signal(timeout=0.2)
        self.assertEqual(sig, ("context1", "pub1", "sig3", (13, "C"), 1))
        self.assertFalse(recv1.has_signal_ready())

        # Unsubscribe the remaining receiver, then immediately resubscribe.
        proxy.sig3.unsubscribe(recv2)
        proxy.sig3.subscribe(recv1)

        # Check that signals are still correctly delivered.
        pub1.send3(14, "D")
        sig = recv1.get_next_signal(timeout=0.2)
        self.assertEqual(sig, ("context1", "pub1", "sig3", (14, "D"), 2))
        self.assertFalse(recv2.has_signal_ready())


class TestRemotePubSubQmi(unittest.TestCase):
    """Test publish/subscribe between contexts, created with qmi.start() instead of QMI_Context()."""

    def setUp(self):

        # Initialize two QMI contexts. Use random numbers to avoid conflicts on server between contexts.
        random_port = random.randint(10000, 30000)  # dynamic port numbers
        random_context = random.randint(3, 100)
        self.random_context_1 = f"context{random_context}"
        self.random_context_2 = f"context{random_context + 1}"
        context_config = {self.random_context_1: {"tcp_server_port": random_port},
                  self.random_context_2: {"tcp_server_port": random_port + 1}}
        config = CfgQmi(
            contexts={
                self.random_context_1: CfgContext(tcp_server_port=random_port),
                self.random_context_2: CfgContext()
            }
        )
        self.context3 = QMI_Context(self.random_context_1, config)
        self.context3.start()
        qmi.start(self.random_context_2, context_cfg=context_config)
        # Create publisher in context3.
        self.pub1 = self.context3.make_rpc_object("pub1", MyPublisher)

    def tearDown(self):
        self.context3.stop()
        qmi.stop()

    def test_pubsub_qmi_get(self):
        """Publish and receive signals between contexts."""
        # The "local" context started in setUp finds and connects to self.context3.
        context = qmi.context().discover_peer_contexts(context_name_filter="context*")
        for ctx in context:
            if ctx[0] == self.random_context_1:
                break

        proxy = qmi.get_rpc_object(f"{self.random_context_1}.pub1", auto_connect=True, host_port=ctx[1])
        # Check that context4 can see the signals from context3.
        sigs = proxy.get_signals()
        expect_sigs = {
            SignalDescription("sig1", ()),
            SignalDescription("sig2", (int,)),
            SignalDescription("sig3", (int, str))
        }
        self.assertEqual(set(sigs), expect_sigs)

        # Create signal receiver queue.
        recv1 = QMI_SignalReceiver()
        recv2 = QMI_SignalReceiver()
        recv3 = QMI_SignalReceiver()

        # Subscribe to signals.
        proxy.sig1.subscribe(recv1)
        proxy.sig2.subscribe(recv2)
        proxy.sig3.subscribe(recv3)

        self.pub1.send1()
        self.pub1.send2()
        self.pub1.send3(12, "C")

        # Wait until remote signals are delivered (if any).
        for _ in range(100):
            if not (recv1.has_signal_ready() and recv2.has_signal_ready() and recv3.has_signal_ready()):
                time.sleep(0.1)

            else:
                break

        # Check that signals are received.
        self.assertTrue(recv1.has_signal_ready())
        self.assertEqual(recv1.get_queue_length(), 1)
        sig = recv1.get_next_signal()
        self.assertEqual(sig, (self.random_context_1, "pub1", "sig1", (), 0))

        self.assertTrue(recv2.has_signal_ready())
        self.assertEqual(recv2.get_queue_length(), 1)
        sig = recv2.get_next_signal()
        self.assertEqual(sig, (self.random_context_1, "pub1", "sig2", (24,), 0))

        self.assertTrue(recv3.has_signal_ready())
        self.assertEqual(recv3.get_queue_length(), 1)
        sig = recv3.get_next_signal()
        self.assertEqual(sig, (self.random_context_1, "pub1", "sig3", (12, "C"), 0))

        # Unsubscribe from remote signal.
        proxy.sig1.unsubscribe(recv1)

        # Publish signals.
        self.pub1.send2()
        self.pub1.send3(22, "D")

        # Wait until remote signals are delivered (if any).
        for _ in range(100):
            if not (recv2.has_signal_ready() and recv3.has_signal_ready()):
                time.sleep(0.1)

            else:
                break

        # Check that local signals are received for receivers 2 and 3, but not for 1.
        self.assertFalse(recv1.has_signal_ready())
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException):
            recv1.get_next_signal()

        self.assertTrue(recv2.has_signal_ready())
        self.assertEqual(recv2.get_queue_length(), 1)
        sig = recv2.get_next_signal()
        self.assertEqual(sig, (self.random_context_1, "pub1", "sig2", (24,), 1))

        self.assertTrue(recv3.has_signal_ready())
        self.assertEqual(recv3.get_queue_length(), 1)
        sig = recv3.get_next_signal()
        self.assertEqual(sig, (self.random_context_1, "pub1", "sig3", (22, "D"), 1))


if __name__ == "__main__":
    unittest.main()



"""
Physik Instrumente (PI) GmbH, servo controller.

The qmi.instruments.picoquant package provides support for:
- E-873 servo controller.
"""

from qmi.instruments.pi.e873 import PI_E873, ReferenceTarget, ReferenceSignalMode



"""Instrument driver for Boston Micromachines Multi-DM series deformable mirror.

This driver depends on a custom helper program to access the deformable mirror.
The helper program is written in C and links with the Boston Micromachines
libraries to access the mirror via USB.

See `qmi/toolage/boston_micromachines_multidm/` for the source code of this helper program.
"""

import logging
import os
import os.path
import re
import subprocess

from typing import List

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_UsageException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


def read_mirror_shape(file_name: str) -> List[float]:
    """Read a mirror shape from a file.

    This function expects the file in the same format used by the
    Boston Micromachines software: a plain text file containing one line
    per actuator, each line containing a single floating point number.

    Parameters:
        file_name: Name of the file to read.

    Returns:
        List of actuator values in the range 0.0 to 1.0.
    """
    shape: List[float] = []
    with open(file_name, "r") as file_handle:
        while True:
            s = file_handle.readline()
            if not s:
                break
            v = float(s.strip())
            shape.append(v)
    return shape


def write_mirror_shape(file_name: str, shape: List[float]) -> None:
    """Write mirror shape to a file.

    Parameters:
        file_name: Name of file to write.
        shape: List of 140 floating point values describing the displacement of each actuator
            as a value in the range 0.0 to 1.0.
    """
    with open(file_name, "w") as file_handle:
        for shape_val in shape:
            print("{:.9f}".format(shape_val), file=file_handle)


class BostonMicromachines_MultiDM(QMI_Instrument):
    """Instrument driver for the Boston Micromachines Multi-DM series deformable mirrors.

    This driver was developed for the DM140A, but is expected to also work
    with other Multi-DM models with 140 actuators.

    This drives communicates with the deformable mirror via an external
    helper program `boston_micromachines_multidm_set_shape`.
    """

    # Number of actuators in the deformable mirror.
    NUMBER_OF_ACTUATORS = 140

    # Expected length of the serial number string.
    BMC_SERIAL_NUMBER_LEN = 11

    # Timeout for running the set_shape helper program (seconds).
    SET_SHAPE_TIMEOUT = 10

    def __init__(self, context: QMI_Context, name: str, serial_number: str, set_shape_prog: str) -> None:
        """Initialize the driver.

        Parameters:
            name: Name for this instrument instance.
            serial_number: Serial number of the deformable mirror.
            set_shape_prog: Location of helper program to load a mirror shape.

        Raises:
            QMI_InstrumentException: By invalid serial number length or string.
        """

        if len(serial_number) != self.BMC_SERIAL_NUMBER_LEN:
            raise QMI_InstrumentException("Invalid serial number {!r}, expecting {} characters"
                                          .format(serial_number, self.BMC_SERIAL_NUMBER_LEN))

        if not re.match(r"^[-0-9A-Za-z#]+$", serial_number):
            raise QMI_InstrumentException("Invalid serial number {!r}".format(serial_number))

        super().__init__(context, name)
        self._serial_number = serial_number
        self._set_shape_prog = set_shape_prog

    def _run_set_shape_prog(self, shape: List[float]) -> None:
        """Run the helper program to load the specified mirror shape."""

        # Serialize the mirror shape to a string for the helper program.
        # One line per actuator holding a floatping point value.
        helper_input = b"".join([(b"%.9f\n" % shape_val) for shape_val in shape ])

        # Prepare command line to start the helper program.
        args = [self._set_shape_prog, "-s", self._serial_number, "-i"]

        _logger.info("[%s] Running helper program %s", self._name, " ".join([repr(arg) for arg in args]))

        # Start the helper program.
        proc = subprocess.Popen(args=args,
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE)

        assert proc.stdin is not None
        assert proc.stdout is not None

        try:

            # Provide input to helper program and capture output from program.
            (helper_output, _) = proc.communicate(input=helper_input, timeout=self.SET_SHAPE_TIMEOUT)

            # Check for error messages in the output from the program.
            output_str = helper_output.decode('latin1')
            for output_line in output_str.splitlines():
                _logger.debug("helper: %s", output_line)

                # Any line starting with "ERROR" indicates an irrecoverable error.
                if output_line.startswith("ERROR"):
                    raise QMI_InstrumentException("Helper program failed: {}".format(output_line))

            # Check exit status of helper program.
            if proc.returncode != 0:
                raise QMI_InstrumentException("Helper program returned exit status {}".format(proc.returncode))

        finally:
            # Whatever happened (success or failure), we need to make sure
            # the program stops and input/output redirection is cleaned up.
            proc.kill()

            proc.stdin.close()
            proc.wait()
            proc.stdout.close()

        _logger.debug("[%s] Helper program successful", self._name)

    @rpc_method
    def open(self) -> None:

        # Nothing actually needs to be done to "open" the mirror.
        # There is no sense in which we can access the mirror device until
        # the point where we activate a mirror shape.
        #
        # Instead of "opening" the device, we just check that the
        # helper program is installed on the system.

        if not os.path.isfile(self._set_shape_prog):
            raise QMI_InstrumentException("Helper program {!r} not found".format(self._set_shape_prog))

        super().open()

    @rpc_method
    def apply(self, shape: List[float]) -> None:
        """Apply actuator values.

        Parameters:
            shape: List of 140 floating point values describing the displacement of each actuator
                as a value in the range 0.0 to 1.0.

        Note that the shape values represent the raw, uncalibrated values sent to the mirror actuators.
        The actual shape of the mirror will be affected by offset mismatch and by interaction
        between neighbouring actuators.
        """
        self._check_is_open()

        if len(shape) != self.NUMBER_OF_ACTUATORS:
            raise QMI_UsageException("Expecting {} actuator values".format(self.NUMBER_OF_ACTUATORS))

        if min(shape) < 0.0 or max(shape) > 1.0:
            raise QMI_UsageException("Expecting actuator values between 0.0 and 1.0")

        _logger.debug("[%s] Applying mirror shape", self._name)

        # Run BMCLoadShape to load the mirror shape.
        self._run_set_shape_prog(shape)



import unittest

import numpy as np

from qmi.instruments.picoquant import TttrHistogram


class PicoQuantSomeHarpOpenTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self._tttr_histogram = TttrHistogram(0, 12500)  # register channel #0 events

    def tearDown(self) -> None:
        pass

    def test_process(self):
        """Test that process can collect data if sync is present"""
        # Events are returned as an array of event records.
        # Each event record contains two fields:
        # | `type` (uint8): The channel index where the event was recorded, or 64 for a SYNC event.
        # | `timestamp` (uint64): Event timestamp as a multiple of the instrument base resolution.
        channel_index = [0, 1, 64, 0, 1, 64, 0, 1]  # Two hits
        expected_counts = 2
        expected_deltas = [1, 1]
        timestamp = [2134567890, 2134567891, 2134567892, 2134567893, 2134567894, 2134567895, 2134567896, 2134567897]
        expected_previous = timestamp[5]
        events = np.array(list(zip(*[iter(channel_index)], *[iter(timestamp)])),
                 dtype=[("type", "uint8"), ("timestamp", "uint64")])
        self._tttr_histogram.process(events)

        self.assertEqual(sum(self._tttr_histogram.counts), expected_counts)
        self.assertListEqual(list(self._tttr_histogram.deltas), expected_deltas)
        self.assertEqual(self._tttr_histogram.previous, expected_previous)

    def test_process_no_sync_event(self):
        """Test that process does not collect data if no sync is present"""
        channel_index = [0, 1, 0, 1, 0, 1]  # zero syncs
        expected_counts = 0
        timestamp = [2134567890, 2134567891, 2134567892, 2134567893, 2134567894, 2134567895]
        events = np.array(list(zip(*[iter(channel_index)], *[iter(timestamp)])),
                 dtype=[("type", "uint8"), ("timestamp", "uint64")])
        self._tttr_histogram.process(events)

        self.assertEqual(sum(self._tttr_histogram.counts), expected_counts)
        self.assertIsNone(self._tttr_histogram.deltas)
        self.assertIsNone(self._tttr_histogram.previous)

    def test_process_append(self):
        """Test that process appends data at second call"""
        # Events are returned as an array of event records.
        # Each event record contains two fields:
        # | `type` (uint8): The channel index where the event was recorded, or 64 for a SYNC event.
        # | `timestamp` (uint64): Event timestamp as a multiple of the instrument base resolution.
        channel_index = [0, 1, 64, 0, 1, 64, 0, 1]  # Two hits for first, three for second
        expected_counts = 5
        expected_deltas = [3, 1, 1]
        timestamp = [2134567890, 2134567891, 2134567892, 2134567893, 2134567894, 2134567895, 2134567896, 2134567897]
        events = np.array(list(zip(*[iter(channel_index)], *[iter(timestamp)])),
                 dtype=[("type", "uint8"), ("timestamp", "uint64")])
        self._tttr_histogram.process(events)

        timestamp2 = [2134567898, 2134567899, 2134567900, 2134567901, 2134567902, 2134567903, 2134567904, 2134567905]
        expected_previous = timestamp2[5]
        events = np.array(list(zip(*[iter(channel_index)], *[iter(timestamp2)])),
                 dtype=[("type", "uint8"), ("timestamp", "uint64")])
        self._tttr_histogram.process(events)
        self.assertEqual(sum(self._tttr_histogram.counts), expected_counts)
        self.assertListEqual(list(self._tttr_histogram.deltas), expected_deltas)
        self.assertEqual(self._tttr_histogram.previous, expected_previous)

    def test_reset(self):
        """Test that reset indeed nullifies values."""
        channel_index = [0, 1, 64, 0, 1, 64, 0, 1]  # Two hits
        expected_counts = 0
        timestamp = [2134567890, 2134567891, 2134567892, 2134567893, 2134567894, 2134567895, 2134567896, 2134567897]
        events = np.array(list(zip(*[iter(channel_index)], *[iter(timestamp)])),
                 dtype=[("type", "uint8"), ("timestamp", "uint64")])
        self._tttr_histogram.process(events)

        self._tttr_histogram.reset()
        self.assertEqual(sum(self._tttr_histogram.counts), expected_counts)
        self.assertIsNone(self._tttr_histogram.deltas)


if __name__ == '__main__':
    unittest.main()



"""Debugging aid."""

import logging
import threading
import time
from collections import namedtuple
from typing import Any, Dict, Optional

_logger = logging.getLogger(__name__)

ObjectRegistration = namedtuple("ObjectRegistration", "oid, timestamp, obj_str, obj_repr, type_str, comment")


class ObjectRegistry:

    def __init__(self) -> None:
        self._mutex = threading.Lock()
        self._counter = 0
        self._registry: Dict[int, ObjectRegistration] = {}

    def register(self, obj: Any, comment: Optional[str] = None) -> int:
        timestamp = time.time()
        with self._mutex:
            oid = self._counter
            self._counter += 1
            registration = ObjectRegistration(oid, timestamp, str(obj), repr(obj), str(type(obj)), comment)
            self._registry[oid] = registration
        return oid

    def unregister(self, oid: int) -> None:
        err_flag = False
        with self._mutex:
            if oid in self._registry:
                del self._registry[oid]
            else:
                err_flag = True

        if err_flag:
            _logger.error("Attempt to unregister object that is not registered: {}".format(oid))

    def report(self, force_summary_flag: bool = True) -> None:

        # Make a copy of the registry and counter.
        with self._mutex:
            registry = self._registry.copy()
            counter = self._counter

        if force_summary_flag or len(registry) > 0:
            num_past_objects = counter - len(registry)
            _logger.info("Number of objects currently in registry: {} (properly registered/unregistered: {}).".format(len(registry), num_past_objects))

        for registration in registry.values():
            _logger.info("registered object: {}".format(registration))



"""Instrument driver for the Bristol 871A Laser Wavelength Meter."""

import collections
import logging
import math
import struct
import time
from typing import List, NamedTuple, Optional

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.thread import QMI_Thread
from qmi.core.transport import create_transport, QMI_Transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)

# Wavelength measurements produced by instrument.
Measurement = NamedTuple(
    "Measurement",
    [
        ("timestamp", float),
        ("index", int),
        ("status", int),
        ("wavelength", float),
        ("power", float),
    ],
)


class _ReaderThread(QMI_Thread):
    """Background thread to read streaming samples from the instrument's RS-422 port."""

    # Check for thread shutdown request every POLL_DURATION seconds.
    POLL_DURATION = 0.100

    def __init__(self, transport: QMI_Transport, queue: collections.deque) -> None:
        # A more specific type annotation would be "queue: Deque[Measurement]".
        # However we can not use it because "Deque" does not exist in Python 3.5.3.
        super().__init__()
        self._transport = transport
        self._queue = queue

    def run(self) -> None:
        """Read measurements from serial port and append to internal queue."""

        while not self._shutdown_requested:
            # Read next measurement from serial stream.
            measurement = self._read_measurement()

            if measurement is not None:
                # Append measurement to queue.
                # NOTE: Appending to a deque is thread-safe.
                self._queue.append(measurement)

    def _read_measurement(self) -> Optional[Measurement]:
        """Read a single measurement from the serial port.

        Returns:
            Next measurement, or None if shutdown requested before a complete measurement was received.
        """

        # Skip until start-of-measurement byte.
        while not self._shutdown_requested:
            data = self._transport.read_until_timeout(1, timeout=self.POLL_DURATION)
            if data == b"\x7e":
                # Got start-of-measurement byte.
                break

        if self._shutdown_requested:
            return None

        # Take timestamp of measurement.
        timestamp = time.time()

        # Receive complete measurement (20 bytes after decoding).
        message_length = 20
        buf = bytearray()
        escape_pending = False
        while (not self._shutdown_requested) and (len(buf) < message_length):
            data = self._transport.read_until_timeout(
                message_length - len(buf), timeout=self.POLL_DURATION
            )
            position = data.find(0x7E)
            if position >= 0:
                # Got another start-of-measurement byte (data corruption).
                buf.clear()
                position += 1
                escape_pending = False
            else:
                position = 0
            while position < len(data):
                if escape_pending:
                    # Decode escaped byte.
                    buf.append(data[position] ^ 0x20)
                    position += 1
                    escape_pending = False
                else:
                    # Find next escape sequence.
                    escape_sequence_position = data.find(0x7D, position)
                    if escape_sequence_position < 0:
                        # Copy rest of string.
                        buf.extend(data[position:])
                        break

                    # Copy until next escape sequence.
                    buf.extend(data[position:escape_sequence_position])
                    position = escape_sequence_position + 1
                    escape_pending = True

        if self._shutdown_requested:
            return None

        # Unpack measurement values:
        #   wavelength: double
        #   power: float
        #   status: uint32
        #   index: uint32
        assert len(buf) == message_length
        (wavelength, power, status, index) = struct.unpack("<dfII", buf)

        # Erroneous measurements are signified by the 'status', but also by the
        # (observed, undocumented) fact that the 'wavelength' is returned as 0.0.
        # To prevent accidents, we replace this error value by NaN.
        if wavelength == 0.0:
            wavelength = math.nan

        return Measurement(timestamp, index, status, wavelength, power)


class Bristol_871A(QMI_Instrument):
    """Instrument driver for the Bristol 871A wavelength meter.

    This driver implements two connections to the instrument:

    - A text-based (SCPI) bidirectional channel via TCP port 23, for configuration and on-demand measurements;
    - A binary output channel (RS-422), emitting small packets corresponding to processed measurements.

    Either of the channels (but not both) may be left unspecified at instantiation time.

    Attributes:
        CONDITION_BITS: A dictionary mapping condition values to messages.
        STATUS_BITS: A dictionary mapping status bit indices to a descriptive message.
        DEFAULT_QUEUE_SIZE: An integer that indicates the size of the RS-422 message queue, if no value is specified
            in the constructor.
        RESPONSE_TIMEOUT: The timeout for SCPI commands, in seconds.
        STATUS_MASK: Bitmask value of all status bits that have a defined meaning.
        STATUS_GOOD: Status value of a wavelength measurement without any issues.
    """

    _rpc_constants = ["CONDITION_BITS", "STATUS_BITS", "STATUS_MASK", "STATUS_GOOD"]

    # Condition codes returned by get_condition(), by bit index.
    CONDITION_BITS = {
        0: "The wavelength has already been read for the current scan",
        2: "The previous requested calibration has failed",
        3: "The power value is outside the valid range of the instrument",
        4: "The temperature value is outside the valid rang of the instrument",
        5: "The wavelength value is outside the valid range of the instrument",
        6: "The detector signal is saturated",
        7: "The detector signal is low",
        8: "No measurable signal was detected",
        9: "The pressure value is outside the valid range of the instrument",
        10: "At least one bit is set in the questionable hardware condition register",
    }

    # Status codes returned in the Measurement.status field, by bit index.
    STATUS_BITS = {
        2: "Reference laser locked",
        3: "Etalon fringe error",
        4: "Etalon saturation error",
        7: "Calibration failure",
        11: "Reference laser not stable",
        13: "Temperature high",
        14: "Temperature low",
        15: "Pressure high",
        16: "Pressure low",
        17: "Wavelength outside instrument specification",
        19: "Etalon fringe error",
        20: "Etalon saturation error",
        21: "Calibration in progress",
        22: "Etalon fringe error",
        23: "Etalon saturation error",
        27: "Wavelength accuracy reduced",
        29: "Low detector signal",
        30: "Fringe frequency error",
    }

    # Default queue holds up to 10000 measurements.
    DEFAULT_QUEUE_SIZE = 10000

    # We expect the instrument to respond within 2 seconds.
    RESPONSE_TIMEOUT = 2.0

    # Status bit mask, indicating all bits that have a well-defined meaning.
    STATUS_MASK = 0x68FBE89C

    # Expected status word for correct, accurate samples.
    STATUS_GOOD = 0x00000004

    def __init__(
        self,
        context: QMI_Context,
        name: str,
        scpi_transport: Optional[str],
        serial_transport: Optional[str],
        queue_size: int = DEFAULT_QUEUE_SIZE,
    ) -> None:
        """Initialize the instrument driver.

        Args:
            name: Name for this instrument instance.
            scpi_transport: QMI transport descriptor for the SCPI channel.
                If not specified, the SCPI channel will not be opened.
            serial_transport: QMI transport descriptor for the serial output channel.
                If not specified, the SCPI channel will not be opened.

        Raises:
            ValueError: At least one of scpi_transport or serial_transport must be specified.
            QMI_TransportDescriptorException: a bad transport string was specified.
        """
        super().__init__(context, name)

        if scpi_transport is None and serial_transport is None:
            raise ValueError(
                "Either scpi_transport or serial_transport should be specified."
            )

        self._scpi_transport = None  # type: Optional[QMI_Transport]
        self._scpi_protocol = None  # type: Optional[ScpiProtocol]
        self._serial_transport = None  # type: Optional[QMI_Transport]
        self._reader_thread = None  # type: Optional[_ReaderThread]
        self._reader_queue = collections.deque(
            maxlen=queue_size
        )  # type: collections.deque

        if scpi_transport is not None:
            self._scpi_transport = create_transport(scpi_transport)
            self._scpi_protocol = ScpiProtocol(self._scpi_transport)

        if serial_transport is not None:
            self._serial_transport = create_transport(serial_transport)
            self._reader_thread = _ReaderThread(
                self._serial_transport, self._reader_queue
            )
            self._reader_thread.start()

    @rpc_method
    def open(self) -> None:
        """Open the SCPI and/or serial channel to the wavemeter."""
        _logger.info("Opening connection to instrument")
        if self._scpi_transport is not None:
            self._scpi_transport.open()
        if self._serial_transport is not None:
            self._serial_transport.open()
        super().open()
        if self._scpi_transport is not None:
            try:
                self._scpi_handshake()
            except Exception:
                self._scpi_transport.close()
                raise

    @rpc_method
    def close(self) -> None:
        """Close the SCPI and/or serial channel to the wavemeter."""
        _logger.info("Closing connection to instrument")
        self._check_is_open()
        if self._reader_thread is not None:
            self._reader_thread.shutdown()
            self._reader_thread.join()
        super().close()
        if self._serial_transport is not None:
            self._serial_transport.close()
        if self._scpi_transport is not None:
            self._scpi_transport.close()

    def _write_scpi(self, cmd: str) -> None:
        """Send SCPI command to instrument."""

        assert self._scpi_protocol is not None
        self._scpi_protocol.write(cmd)

    def _ask_scpi(self, cmd: str) -> str:
        """Send command to instrument and return response from instrument."""

        assert self._scpi_protocol is not None
        return self._scpi_protocol.ask(cmd).rstrip("\r\n")

    @rpc_method
    def reset(self) -> None:
        """Reset the wavemeter using the SCPI '\\*RST' command.

        This resets (most) settings to their default values.
        """
        _logger.info("Resetting %s", self._name)
        self._write_scpi("*RST")

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read wavemeter instrument type and version

        Returns:
            A QMI_InstrumentIdentification instance.
        """
        resp = self._ask_scpi("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException(
                "Unexpected response to *IDN?, got {!r}".format(resp)
            )
        return QMI_InstrumentIdentification(
            vendor=words[0].strip(),
            model=words[1].strip(),
            serial=words[2].strip(),
            version=words[3].strip(),
        )

    def _scpi_handshake(self) -> None:
        """Read initial welcome message from Telnet port."""

        assert self._scpi_transport is not None

        while True:
            resp = self._scpi_transport.read_until(
                message_terminator=b"\n", timeout=self.RESPONSE_TIMEOUT
            )
            if resp.find(b"no connections available") >= 0:
                _logger.error("Can not connect to %s SCPI port (%r)", self._name, resp)
                raise QMI_InstrumentException(
                    "Can not connect to {} SCPI port".format(self._name)
                )
            if resp.find(b"Bristol Instruments") >= 0:
                # This is almost the last line of the welcome message. One more empty line follows.
                break

        # Send IDN command.
        self._write_scpi("*IDN?")

        # Wait for IDN response.
        while True:
            resp = self._scpi_transport.read_until(
                message_terminator=b"\n", timeout=self.RESPONSE_TIMEOUT
            )
            if resp.startswith(b"*IDN?"):
                # Got command echo. This should never happen.
                raise QMI_InstrumentException(
                    "Instrument {} sent unexpected command echo".format(self._name)
                )
            if resp.startswith(b"BRISTOL"):
                # Got IDN response.
                break

    @staticmethod
    def _parse_int(resp: str, cmd: str) -> int:
        """Parse integer response from instrument.

        Return integer value or raise QMI_InstrumentException.
        """
        try:
            return int(resp)
        except ValueError:
            raise QMI_InstrumentException(
                "Expecting integer response to command {!r} but got {!r}".format(
                    cmd, resp
                )
            )

    @staticmethod
    def is_valid_measurement(measurement: Measurement) -> bool:
        """Return True if the specified measurement represents a valid, accurate measured wavelength."""
        return (
            (measurement.status & Bristol_871A.STATUS_MASK) == Bristol_871A.STATUS_GOOD
        ) and (measurement.wavelength > 0)

    @rpc_method
    def read_measurement(self) -> Measurement:
        """Read next measurement and return measured value as a Measurement instance."""
        resp = self._ask_scpi(":READ:ALL?")
        timestamp = time.time()
        words = resp.split(",")
        if len(words) != 4:
            raise QMI_InstrumentException(
                "Unexpected response to :READ:ALL? ({!r})".format(resp)
            )
        try:
            index = int(words[0])
            status = int(words[1])
            wavelength = float(words[2])
            power = float(words[3])
        except ValueError:
            raise QMI_InstrumentException(
                "Unexpected response to :READ:ALL? ({!r})".format(resp)
            )
        # Erroneous measurements are signified by the 'status', but also by the
        # (observed, undocumented) fact that the 'wavelength' is returned as 0.0.
        # To prevent accidents, we replace this error value by NaN.
        if wavelength == 0.0:
            wavelength = math.nan
        return Measurement(timestamp, index, status, wavelength, power)

    @rpc_method
    def calibrate(self) -> None:
        """Start calibration of the instrument.

        Calibration takes approximately 2 seconds and interrupts measurements for its duration.
        """
        self._write_scpi(":SENS:CALI")

        # The calibration takes a few seconds and we like to ensure that the process has finished once we return from
        # this function.
        time.sleep(5.0)

    @rpc_method
    def get_auto_calibration_method(self) -> str:
        """Return the current auto-calibration method.

        :return: "OFF" if auto-calibration is disabled;
            "TIME" when time-driven auto-calibration enabled;
            "TEMP" when temperature-driven auto-calibration enabled.
        """
        return self._ask_scpi(":SENS:CALI:METH?")

    @rpc_method
    def set_auto_calibration_method(self, method: str) -> None:
        """Set auto-calibration method.

        :argument method: Valid values are "OFF" or "TIME" or "TEMP".
        """
        method_upper = method.upper()
        if method_upper not in ("OFF", "TIME", "TEMP"):
            raise ValueError("Invalid auto-calibration method {!r}".format(method))
        self._write_scpi(":SENS:CALI:METH %s" % method_upper)

    @rpc_method
    def get_auto_calibration_temperature(self) -> int:
        """Return the temperature change (in units of 0.1 degree Celsius) that will initiate recalibration."""
        cmd = ":SENS:CALI:TEMP?"
        resp = self._ask_scpi(cmd)
        return self._parse_int(resp, cmd)

    @rpc_method
    def set_auto_calibration_temperature(self, delta: int) -> None:
        """Set the temperature change that will initiate recalibration when temperature-driven calibration is enabled.

        :argument delta: Temperature change that will initiate recalibration (in units of 0.1 degree Celcius).
            Valid range: 1 to 50.
        """
        if delta < 1 or delta > 50:
            raise ValueError("Invalid temperature threshold for auto-calibration")
        self._write_scpi(":SENS:CALI:TEMP %d" % delta)

    @rpc_method
    def get_auto_calibration_time(self) -> int:
        """Return the time interval [minutes] for automatic recalibration."""
        cmd = ":SENS:CALI:TIM?"
        resp = self._ask_scpi(cmd)
        return self._parse_int(resp, cmd)

    @rpc_method
    def set_auto_calibration_time(self, interval: int) -> None:
        """Set the time interval for recalibration (when time-driven calibration is enabled).

        :argument interval: Recalibration interval in minutes. Valid range: 5 to 1440.
        """
        if interval < 5 or interval > 1440:
            raise ValueError("Invalid time interval for auto-calibration")
        self._write_scpi(":SENS:CALI:TIM %d" % interval)

    @rpc_method
    def get_condition(self) -> int:
        """Read questionable condition register."""
        cmd = ":STAT:QUES:COND?"
        resp = self._ask_scpi(cmd)
        return self._parse_int(resp, cmd)

    @rpc_method
    def get_trigger_method(self) -> str:
        """Return the current trigger method.

        :return: Trigger method: "INT" for internal triggering;
            "FALL" for external trigger on falling edge;
            "RISE" for external trigger on rising edge.
        """
        return self._ask_scpi(":TRIG:SEQ:METH?")

    @rpc_method
    def set_trigger_method(self, method: str) -> None:
        """Set the trigger method.

        :argument method: Trigger method. Valid values: "INT", "FALL", "RISE".
        """
        method_upper = method.upper()
        if method_upper not in ("INT", "FALL", "RISE"):
            raise ValueError("Invalid trigger setting")
        self._write_scpi(":TRIG:SEQ:METH %s" % method_upper)

    @rpc_method
    def get_trigger_rate(self) -> int:
        """Return current trigger rate (in Hz) for the internal trigger."""
        cmd = ":TRIG:SEQ:RATE:ADJ?"
        resp = self._ask_scpi(cmd)
        return self._parse_int(resp, cmd)

    @rpc_method
    def set_trigger_rate(self, rate: int) -> None:
        """Set trigger rate for the internal trigger.

        Args:
            rate: Trigger rate in Hz, or 0 to adjust to optical illumination.
                Valid values are 20, 50, 100, 250, 500, 1000 and 0.
        """
        if rate not in (0, 20, 50, 100, 250, 500, 1000):
            raise ValueError("Invalid trigger rate")
        if rate == 0:
            self._write_scpi(":TRIG:SEQ:RATE:ADJ")
        else:
            self._write_scpi(":TRIG:SEQ:RATE %d" % rate)

    @rpc_method
    def memory_start(self) -> None:
        """Start recording samples in internal memory in the instrument.

        KNOWN FIRMWARE ISSUE: The MMEM INIT/OPEN/CLOSE/DATA? sequence often stops producing samples after running for
            some hours.
        """
        self._write_scpi(":MMEM:INIT")
        self._write_scpi(":MMEM:OPEN")

    @rpc_method
    def memory_stop(self) -> None:
        """Stop recording samples in internal memory.

        KNOWN FIRMWARE ISSUE: The MMEM INIT/OPEN/CLOSE/DATA? sequence often stops producing samples after running for
            some hours.
        """
        self._write_scpi(":MMEM:CLOSE")

    @rpc_method
    def get_memory_contents(self) -> List[Measurement]:
        """Fetch recorded samples from internal memory.

        KNOWN FIRMWARE ISSUE: The MMEM INIT/OPEN/CLOSE/DATA? sequence often stops producing samples after running for
            some hours.
        """

        self._write_scpi(":MMEM:DATA?")
        assert self._scpi_protocol is not None
        data = self._scpi_protocol.read_binary_data()

        message_length = 20
        nbytes = len(data)
        if nbytes % message_length != 0:
            raise QMI_InstrumentException(
                "Expected multiple of 20 bytes in memory but got {} bytes".format(
                    nbytes
                )
            )
        nsamples = nbytes // message_length

        # Unpack measurement values:
        #   wavelength: double
        #   power: float
        #   status: uint32
        #   index: uint32
        ret = []  # type: List[Measurement]
        timestamp = time.time()
        for i in range(nsamples):
            sample = data[i * message_length : (i + 1) * message_length]
            (wavelength, power, status, index) = struct.unpack("<dfII", sample)
            # Erroneous measurements are signified by the 'status', but also by the
            # (observed, undocumented) fact that the 'wavelength' is returned as 0.0.
            # To prevent accidents, we replace this error value by NaN.
            if wavelength == 0.0:
                wavelength = math.nan
            ret.append(Measurement(timestamp, index, status, wavelength, power))

        return ret

    @rpc_method
    def get_streaming_measurements(self) -> List[Measurement]:
        """Return streaming measurements received from the instrument.

        If the instrument uses a serial transport (in addition to the SCPI transport),
        streaming measurements are received from the instrument and buffered in
        the driver. This method returns the list of measurements currently
        in the buffer and removes them from the buffer.

        The buffer has a maximum size of STREAMING_QUEUE_SIZE measurements.
        If the buffer becomes full, old measurements are dropped automatically.

        Returns:
            A list of measurements received since last call.
        """
        ret = []
        for _ in range(len(self._reader_queue)):
            ret.append(self._reader_queue.popleft())
        return ret



"""
Instrument driver for the PicoTech PicoScope 3000A series oscilloscopes.

This driver depends on the PicoSDK libraries from PicoTech and on
the PicoSDK Python wrappers from
https://github.com/picotech/picosdk-python-wrappers.
"""

import ctypes
import logging
from typing import TYPE_CHECKING
import warnings

import numpy as np
from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_UsageException
from qmi.core.rpc import rpc_method

from qmi.instruments.picotech import ChannelCoupling, TriggerEdge
from qmi.instruments.picotech._picoscope import PicoTech_PicoScope, _check_error

# Lazy import of the "picosdk" module. See the function _import_modules() _picoscope.py.
if TYPE_CHECKING:
    from picosdk.ps3000a import ps3000a as _ps
else:
    _ps = None

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class PicoTech_PicoScope3403(PicoTech_PicoScope):
    """Instrument driver for the PicoTech PicoScope 3403 USB oscilloscope.

    Attributes:
        NUM_CHANNELS:     Number of oscilloscope channels.
        NUM_INPUT_RANGES: Number of supported input ranges. Range '0' is not supported.
    """
    NUM_CHANNELS = 4
    NUM_INPUT_RANGES = 11

    def __init__(self, context: QMI_Context, name: str, serial_number: str) -> None:
        """Initialize the instrument driver.

        Parameters:
            name:          Name for this instrument instance.
            serial_number: Serial number of the Picoscope.
        """
        super().__init__(context, name, serial_number)
        self._library = "3000a"

    def _get_time_base(self, sampling_interval: int) -> int:
        """Returns the time base selector for a given sampling interval.
        For invalid sampling intervals, an exception is raised.

        Parameters:
            sampling_interval: Sampling interval in nanoseconds. Must be 1, 2, 4 or a multiple of 8 ns.

        Returns:
            time_base:         The respective time-base value for the sampling interval.
        """
        if sampling_interval in (1, 2, 4):
            return int(np.log2(sampling_interval))

        if sampling_interval > 4 and sampling_interval % 8 == 0:
            return int(round((sampling_interval / 8) + 2, 0))

        raise QMI_UsageException("Sampling interval must be 1, 2, 4 or a multiple of 8 ns")

    @rpc_method
    def run_block(self, num_pretrig_samples: int, num_posttrig_samples: int, time_base: int) -> None:
        """Start acquisition in block mode.

        If the trigger is enabled (see `set_trigger()`), the oscilloscope will
        until the trigger condition occurs, then acquire a single block of data.
        If the trigger is disabled, the oscilloscope will immediately acquire
        a single block of data.

        This function returns immediately while the acquisition is in progress.
        Call `is_block_ready()` or `wait_block_ready()` to check whether the acquisition has finished.
        Optionally call `stop()` to end the acquisition before it completes.

        Parameters:
            num_pretrig_samples:    Number of pre-trigger samples.
            num_posttrig_samples:   Number of post-trigger samples.
            time_base:              Time base selector (range 0 ... 2**32-1). The effective time base is
                                    (timebase - 2) * 8 ns for timebase > 2, else 2^timebase.
                                    For one enabled channel, values from 0 are allowed (time-base >= 1 ns).
                                    For two enabled channels, values from 1 are allowed (time-base >= 2 ns).
                                    From three enabled channels, values from 2 are allowed (time-base >= 4 ns).
        """
        self._check_is_open()

        num_samples = num_pretrig_samples + num_posttrig_samples
        par_interval_ns = ctypes.c_float()
        err = _ps.ps3000aGetTimebase2(
            self._handle,
            time_base,
            num_samples,
            ctypes.byref(par_interval_ns),
            0,  # oversamples
            None,  # maxSamples
            0,  # segmentIndex
        )
        _check_error(err)

        err = _ps.ps3000aRunBlock(
            self._handle,
            num_pretrig_samples,
            num_posttrig_samples,
            time_base,
            0,  # oversamples
            None,  # timeIndisposedMs
            0,  # segmentIndex
            None,  # lpReady
            None,
            )  # pParameter
        _check_error(err)

        self._num_samples = num_samples
        self._timebase_interval_ns = par_interval_ns.value

    @rpc_method
    def acquire_by_trigger_and_get_data(
        self,
        channels: list[int],
        max_val: list[float],
        couplings: list[ChannelCoupling],
        trigger_channel: int | None,
        trigger_level: float,
        sampling_interval: int,
        time_span: int,
    ) -> object:
        """Acquires data from a specific channel, which is triggered, and returns the data.

        It is assumed that the trigger level is set to (almost) the highest point of
        the measured waveform. This is used to select the appropriate input range.
        The scope waits 1 second to get triggered.
        Default measurement values are: DC coupling and triggering on the rising edge.
        The number of samples is determined from the sampling interval and time span and set
        relative to the trigger point.
        If no trigger channel number is given, trigger is disabled and sampling is done from
        the number of samples before the current moment of `run_block` call.
        The data is then sent back as real voltages.

        Parameters:
            channels:          List of used channels [0...NUM_CHANNELS-1] for readout.
            max_val:           Maximum expected values for ALL channels [0, 1, ... NUM_CHANNELS-1] to set an
                               appropriate input range.
            couplings:         Coupling of channels for readout for ALL channels [0, 1, ... NUM_CHANNELS-1].
                               ChannelCoupling.DC or ChannelCoupling.AC.
            trigger_channel:   Channel to set the trigger on. If it is 'None', triggers will be disabled on
                               all channels.
            trigger_level:     Threshold in Volts for the trigger.
            sampling_interval: Sampling interval in nanoseconds. Must be 1, 2, 4 or a multiple of 8 ns.
            time_span:         Time span of data acquiring in nanoseconds.

        Returns:
            times:    A 1D Numpy array containing the sampling times in ns.
            voltages: A 1D Numpy array containing the measured voltages in V.
        """
        warnings.warn(
            f"{self.acquire_by_trigger_and_get_data.__name__} will be deprecated in the future as it is not a base "
            "functionality of the driver. Please implement it as a custom method in your code.")

        if min(channels) < 0 or max(channels) >= self.NUM_CHANNELS:
            raise ValueError("Invalid channel index")

        if trigger_channel is not None:
            if (trigger_channel < 0) or (trigger_channel >= self.NUM_CHANNELS) or trigger_channel not in channels:
                raise ValueError("Invalid trigger channel index")

        time_base = self._get_time_base(sampling_interval)
        if time_base < 2 and len(channels) > 2 ** time_base:
            raise QMI_UsageException(f"Time-base of {time_base} is not supported for {len(channels)} channels")

        # Give maximum voltage range per expected maximum value per enabled channel.
        max_volts = []
        for chan in range(self.NUM_CHANNELS):
            if chan not in channels:
                # Disable all channels that are not used and skip to next one.
                self.set_channel(chan, False, ChannelCoupling.DC, 0, 0)
                continue

            if np.abs(max_val[chan]) <= 0.02:
                ranges = 1
                max_volts.append(0.02)
            elif np.abs(max_val[chan]) <= 0.05:
                ranges = 2
                max_volts.append(0.05)
            elif np.abs(max_val[chan]) <= 0.1:
                ranges = 3
                max_volts.append(0.1)
            elif np.abs(max_val[chan]) <= 0.2:
                ranges = 4
                max_volts.append(0.2)
            elif np.abs(max_val[chan]) <= 0.5:
                ranges = 5
                max_volts.append(0.5)
            elif np.abs(max_val[chan]) <= 1:
                ranges = 6
                max_volts.append(1)
            elif np.abs(max_val[chan]) <= 2:
                ranges = 7
                max_volts.append(2)
            elif np.abs(max_val[chan]) <= 5:
                ranges = 8
                max_volts.append(5)
            elif np.abs(max_val[chan]) <= 10:
                ranges = 9
                max_volts.append(10)
            else:
                ranges = 10
                max_volts.append(20)

            self.set_channel(chan, True, couplings[chan], ranges, 0)

        # Calculate the total number of samples in the time span.
        number_samples = time_span // sampling_interval
        _logger.info("Timespan = %i Num samples = %i", time_span, number_samples)

        if trigger_channel is not None:
            self.stop()
            trig_int = int(self.MAX_SAMPLE_VALUE / max_volts[channels.index(trigger_channel)] * trigger_level)
            self.set_trigger(True, trigger_channel, trig_int, TriggerEdge.RISING)
            self.run_block(number_samples // 2, number_samples - number_samples // 2, time_base)

        else:
            self.disable_trigger()
            self.stop()
            self.run_block(number_samples, 0, time_base)

        self.wait_block_ready(timeout=1.5)
        data = self.get_block_data(channels)
        times = np.arange(0, number_samples, 1) * sampling_interval
        voltages = []
        for i in channels:
            voltages.append(data[0][channels.index(i)] * (max_volts[channels.index(i)] / self.MAX_SAMPLE_VALUE))

        return times, voltages

    @rpc_method
    def get_sampling_interval(self, time_base: int) -> float:
        """Returns the scope's time resolution in nanoseconds depending on the time_base selector.

        Parameters:
            time_base:   Timebase selector (range 0 ... 2^32-1).
                         The effective time resolution is 2^time_base ns for timebase <= 2.
                         The effective time resolution is (timebase-2)/(125000000) for timebase > 2.

        Returns:
            resolution:  Scope's time resolution in nanoseconds. 0 is returned for invalid time_base.
        """
        if 0 <= time_base <= 2:
            return 2.0 ** time_base

        elif 2 < time_base < 2 ** 32:
            # We can round here to 0 decimals as this model has always a resolution of multiple of 8.
            return round(((time_base - 2) / 125e6) * 1e9, 0)

        else:
            return 0



import unittest
import unittest.mock

import qmi
from qmi.instruments.agiltron import Agiltron_Ff1x4

import qmi.core.exceptions
from qmi.core.context import QMI_Context


class TestAgiltronFF1x4(unittest.TestCase):

    def setUp(self) -> None:
        unittest.mock.patch("qmi.core.transport.QMI_SerialTransport._validate_device_name")
        qmi_context = unittest.mock.MagicMock(spec=QMI_Context)
        qmi_context.name = "mockers"
        self.ser_address = "COM100"
        self.baudrate = 9600
        self.timeout = 1
        transport_id = "serial:{}:baudrate={}".format(self.ser_address, self.baudrate)
        self.agiltron = Agiltron_Ff1x4(qmi_context, "switchy", transport_id)

    def tearDown(self) -> None:
        self.agiltron.close()

    def test_open_close(self):
        with unittest.mock.patch("serial.Serial") as ser:
            self.agiltron.open()
            ser.assert_called_once_with(
                self.ser_address,
                baudrate=self.baudrate,  # The rest are defaults
                bytesize=8,
                parity='N',
                rtscts=False,
                stopbits=1,
                timeout=0.04
                )

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read")
    def test_get_active_channel(self, mock_read):
        """See that the get_active_channel RPC method returns the expected channel number."""
        expected = 4
        mock_read.return_value = bytes([0x01, 0x02, 0x03, expected])
        with unittest.mock.patch("serial.Serial"):
            self.agiltron.open()
            channel = self.agiltron.get_active_channel()

        self.assertEqual(channel, expected)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read")
    def test_set_channel_active(self, mock_read):
        """See that the set_channel_active RPC method sets the expected channel active."""
        channel = 3
        expected = bytes([0x01, 0x12, 0x00, int(channel)])
        mock_read.return_value = expected
        with unittest.mock.patch("serial.Serial") as ser:
            self.agiltron.open()
            self.agiltron.set_channel_active(channel)
            write_call_in_calls_index = ["().write" in call[0] for call in ser.mock_calls].index(True)
            self.assertTupleEqual(ser.mock_calls[write_call_in_calls_index][1], (expected,))

        self.agiltron._transport._serial.write.assert_called_once_with(expected)

    def test_set_channel_active_with_wrong_channel_number(self):
        """ Test that the channel number checking in set_channel_active works as expected"""
        channels = [0, self.agiltron.CHANNELS + 1]
        with unittest.mock.patch("serial.Serial"):
            self.agiltron.open()
            for channel in channels:
                with self.assertRaises(ValueError):
                    self.agiltron.set_channel_active(channel)

    @unittest.mock.patch("qmi.core.transport.QMI_SerialTransport.read")
    def test_set_channel_active_with_erroneous_reply(self, mock_read):
        """See that exceptions are raised with wrong return data """
        channel = 1
        mock_read.side_effect = [bytes([0x01, 0x12, 0x00, int(channel), 0x05]),  # Test 1
                                 bytes([0x01, 0x12, 0x00, int(channel + 1)]),  # Test 2
                                 bytes([0x01, 0x12, 0x00])]  # Test 3
        # Test 1: too many bytes returned
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException), unittest.mock.patch("serial.Serial"):
            self.agiltron.open()
            self.agiltron.set_channel_active(channel)

        self.agiltron.close()
        # Test 2: correct number of bytes but wrong value
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException), unittest.mock.patch("serial.Serial"):
            self.agiltron.open()
            self.agiltron.set_channel_active(channel)

        self.agiltron.close()
        # Test 3: too few bytes returned
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException), unittest.mock.patch("serial.Serial"):
            self.agiltron.open()
            self.agiltron.set_channel_active(channel + 1)


if __name__ == '__main__':
    unittest.main()
